/*
	Non ASCII characters recognized by STedi for special commands
	The ordering is made to facilitate the internal jumptables, while
	still showing the ASCII origin as much as possible.
	Note that the standard flag system for UNIX systems works here
	almost completely, with the exception of the keys like the
	arrows, home, insert etc.

	On the micro computers some special rewriting may be needed to
	get the keys in this notation.

	The mnemonics that are used here can also be used in keycomp
	programs. In that case they are case insensitive.
	There should be no problems when the numerical entries in the
	rhs should start with a digit and contain at least two characters.
	There are no relevant user flags and masks in the rhs expressions.
	Individual characters at taken at face value, except for:
	\ : ; , and blank which have to be escaped with a \. the alternative
	is to use the mnemonics for them.

	J.Vermaseren 9-Aug-1988.
*/

#define MOUSE		0x70000L

#define ALT_A		0x0141
#define ALT_B		0x0142
#define ALT_C		0x0143
#define ALT_D		0x0144
#define ALT_E		0x0145
#define ALT_F		0x0146
#define ALT_G		0x0147
#define ALT_H		0x0148
#define ALT_I		0x0149
#define ALT_J		0x014A
#define ALT_K		0x014B
#define ALT_L		0x014C
#define ALT_M		0x014D
#define ALT_N		0x014E
#define ALT_O		0x014F
#define ALT_P		0x0150
#define ALT_Q		0x0151
#define ALT_R		0x0152
#define ALT_S		0x0153
#define ALT_T		0x0154
#define ALT_U		0x0155
#define ALT_V		0x0156
#define ALT_W		0x0157
#define ALT_X		0x0158
#define ALT_Y		0x0159
#define ALT_Z		0x015A
#define ALT_1		0x0131
#define ALT_2		0x0132
#define ALT_3		0x0133
#define ALT_4		0x0134
#define ALT_5		0x0135
#define ALT_6		0x0136
#define ALT_7		0x0137
#define ALT_8		0x0138
#define ALT_9		0x0139
#define ALT_0		0x0130
#define ALT_EQ		0x013D
#define CTRL_A		0x0241
#define CTRL_B		0x0242
#define CTRL_C		0x0243
#define CTRL_D		0x0244
#define CTRL_E		0x0245
#define CTRL_F		0x0246
#define CTRL_G		0x0247
#define CTRL_H		0x0248
#define CTRL_I		0x0249
#define CTRL_J		0x024A
#define CTRL_K		0x024B
#define CTRL_L		0x024C
#define CTRL_M		0x024D
#define CTRL_N		0x024E
#define CTRL_O		0x024F
#define CTRL_P		0x0250
#define CTRL_Q		0x0251
#define CTRL_R		0x0252
#define CTRL_S		0x0253
#define CTRL_T		0x0254
#define CTRL_U		0x0255
#define CTRL_V		0x0256
#define CTRL_W		0x0257
#define CTRL_X		0x0258
#define CTRL_Y		0x0259
#define CTRL_Z		0x025A
#define CTRL_1		0x0231
#define CTRL_2		0x0232
#define CTRL_3		0x0233
#define CTRL_4		0x0234
#define CTRL_5		0x0235
#define CTRL_6		0x0236
#define CTRL_7		0x0237
#define CTRL_8		0x0238
#define CTRL_9		0x0239
#define CTRL_0		0x0230
#define CTRL_MIN	0x022D
#define FUN1		0x0401
#define FUN2		0x0402
#define FUN3		0x0403
#define FUN4		0x0404
#define FUN5		0x0405
#define FUN6		0x0406
#define FUN7		0x0407
#define FUN8		0x0408
#define FUN9		0x0409
#define FUN10		0x040A
#define FUN11		0x0801
#define FUN12		0x0802
#define FUN13		0x0803
#define FUN14		0x0804
#define FUN15		0x0805
#define FUN16		0x0806
#define FUN17		0x0807
#define FUN18		0x0808
#define FUN19		0x0809
#define FUN20		0x080A
#define TAG1		0x0C01
#define TAG2		0x0C02
#define TAG3		0x0C03
#define TAG4		0x0C04
#define TAG5		0x0C05
#define TAG6		0x0C06
#define TAG7		0x0C07
#define TAG8		0x0C08
#define TAG9		0x0C09
#define TAG10		0x0C0A
#define UP			0x1000
#define DOWN		0x1001
#define RIGHT		0x1002
#define LEFT		0x1003
#define SH_UP		0x1004
#define SH_DOWN		0x1005
#define SH_RIGHT	0x1006
#define SH_LEFT		0x1007
#define CT_UP		0x1008
#define CT_DOWN		0x1009
#define CT_RIGHT	0x100A
#define CT_LEFT		0x100B
#define SC_UP		0x100C
#define SC_DOWN		0x100D
#define SC_RIGHT	0x100E
#define SC_LEFT		0x100F
#define HOME		0x1010
#define SH_HOME		0x1011
#define CT_HOME		0x1012
#define SC_HOME		0x1013
#define INSERT		0x1014
#define SH_INSERT	0x1015
#define DELETE		0x1016
#define CT_DELETE	0x1017
#define BACKSPACE	0x1018
#define CRETURN		0x1019
#define LINEFEED	0x101A
#define ESCAPE		0x101B
#define UNDO		0x101C
#define HELP		0x101D
#define CT_INSERT	0x101E
#define SC_INSERT	0x101F
#define ENTER		0x1020
#define ESCAPE_C	0x1021
#define P_ITEM1		0x1030
#define P_ITEM2		0x1031
#define P_ITEM3		0x1032
#define P_ITEM4		0x1033
#define P_ITEM5		0x1034
#define P_ITEM6		0x1035
#define P_ITEM7		0x1036
#define P_ITEM8		0x1037
#define P_ITEM9		0x1038
#define P_ITEM10	0x1039
#define TAB			0x0009
#define BLANK		0x0020
#define COMMA		0x002C
#define COLON		0x003A
#define SEMICOLON	0x003B
#define BACKSLASH	0x005C
#define NUM_1		0x2031
#define NUM_2		0x2032
#define NUM_3		0x2033
#define NUM_4		0x2034
#define NUM_5		0x2035
#define NUM_6		0x2036
#define NUM_7		0x2037
#define NUM_8		0x2038
#define NUM_9		0x2039
#define NUM_0		0x2030
#define NUM_LB		0x2028
#define NUM_RB		0x2029
#define NUM_SL		0x202F
#define NUM_ST		0x202A
#define NUM_MI		0x202D
#define NUM_PL		0x202B
#define NUM_PE		0x202E
#define NUM_EN		0x200D
#define M_MEN1		0x4001
#define M_MEN2		0x4002
#define M_MEN3		0x4003
#define M_MEN4		0x4004
#define M_MEN5		0x4005
#define M_MEN6		0x4006
#define M_MEN7		0x4007
#define M_MEN8		0x4008
#define M_MEN9		0x4009
#define M_MEN10		0x400A
#define M_MEN11		0x400B
#define M_MEN12		0x400C
#define M_MEN13		0x400D
#define M_MEN14		0x400E
#define M_MEN15		0x400F
#define M_MEN16		0x4010
#define M_MEN17		0x4011
#define M_MEN18		0x4012
#define M_MEN19		0x4013
#define M_MEN20		0x4014
#define M_MEN21		0x4015
#define M_MEN22		0x4016
#define M_MEN23		0x4017
#define M_MEN24		0x4018
#define M_MEN25		0x4019
#define M_MEN26		0x401A
#define M_MEN27		0x401B
#define M_MEN28		0x401C
#define M_MEN29		0x401D
#define M_MEN30		0x401E
#define M_BLEFT		0x4041		
#define M_BRIGHT	0x4042
#define M_READ		0x4043
#define M_WRITE		0x4044
#define M_UP		0x4045
#define M_DOWN		0x4046
#define M_RIGHT		0x4047
#define M_LEFT		0x4048
#define M_LOOK		0x4049
#define M_ADDWIN	0x404A
#define M_DELWIN	0x404B
#define M_SHOW_D	0x404C
#define M_STORE_D	0x404D
#define M_SET_D		0x404E
#define M_LOAD_D	0x404F
#define M_READFS	0x4050
#define M_WRITEFS	0x4051
#define M_HSLIDER	0x4052
#define M_VSLIDER	0x4053

