/*
  	#[ Includes :
 vi: set ts=4 sw=4:
*/

#include "stedi.h"

#include <stdio.h>
#include <errno.h>

#ifdef BSD
#undef BSD
#endif

#include <termcap.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <termio.h>
struct termio tbufsave;

#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

int winupflag = 0;
int keywaitflag = 0;
int input_fd = 0;
int psbeginfirst = 1;

/*
  	#] Includes :
  	#[ Definitions :
 		#[ Defines :

	Because of problems with `UP' we cannot use keydefs.h here
	Unix must have the variable UP and keydefs.h insists on UP
	being a macro.
	We take here those definitions from keydefs.h that we need
	and modify the names UP, DOWN, LEFT, RIGHT to C_UP etc.
*/

#define C_UP		0x1000L
#define C_DOWN		0x1001L
#define C_LEFT		0x1003L
#define C_RIGHT		0x1002L
#define HOME		0x1010L
#define BACKSPACE	0x1018L
#define ESCAPE		0x101BL
#define DELETE		0x1016L
#define CRETURN		0x1019L
#define LINEFEED	0x101AL
#define SH_UP		0x1004L
#define SH_DOWN		0x1005L
#define SH_LEFT		0x1007L
#define SH_RIGHT	0x1006L
#define CTRL_2		0x0232L
#define INSERT		0x1014L
#define SH_INSERT	0x1015L
#define CT_DELETE	0x1017L
#define UNDO		0x101CL
#define HELP		0x101DL
#define SH_HOME		0x1011L
#define CT_MIN		0x022DL
#define FUN1		0x0401L
#define FUN2		0x0402L
#define FUN3		0x0403L
#define FUN4		0x0404L
#define FUN5		0x0405L
#define FUN6		0x0406L
#define FUN7		0x0407L
#define FUN8		0x0408L
#define FUN9		0x0409L
#define FUN10		0x040AL
#define FUN11		0x0801L
#define FUN12		0x0802L
#define FUN13		0x0803L
#define FUN14		0x0804L
#define FUN15		0x0805L
#define FUN16		0x0806L
#define FUN17		0x0807L
#define FUN18		0x0808L
#define FUN19		0x0809L
#define FUN20		0x080AL

#define TRUE 1
#define FALSE 0

#define MAXCOLS 255
#define MAXPOLLBUF 10

#define BELL '\007'
#define CLRTOEOL putctl(CE)

/*
 		#] Defines :
 		#[ Declarations :
*/
 
VOID setraw(VOID);
VOID unsetraw(VOID);

VOID SetMyBuf(VOID),MyFlush(VOID),UndoMyBuf(VOID);
int OutChar(int);

extern speed_t ospeed;				/* terminal speed (from ioctl()). needed in termcap */
 
WORD ScrollTop = -1;		/* Initialization of the scrolling region */
WORD ScrollBot = -1;
 
typedef struct cellstr {
	UBYTE chr;
	UBYTE att;
} *CELL;

CELL	LogMem;				/* The logical screen address */
CELL	PhysMem;			/* The representation of the physical screen */

UWORD	memsize;
WORD	CurrentAtt;
WORD	CurrentRow;
WORD	CurrentCol;
WORD	WrapMode;
WORD	VisCursor;
WORD	NotNormal = 0;
WORD	AltSet;
 
UBYTE	*ObPoin;
UBYTE	*ObStart = 0;
UBYTE	*ObWrite;

WORD StrCont(BYTE *,BYTE *);
VOID SetCursor(WORD,WORD),SetColor(WORD),putchc(UBYTE),putchs(UBYTE);
void KeyInrupt(int), WinInrupt(int), TstpInrupt(int), ContInrupt(int);

WORD firstKeyIni = 1;
int fcntlflags;
/*
 		#] Declarations :
 		#[ OutPutMacros :	for termcapped output
*/

#define putctl(s) (s != NULL ? tputs(s,1,OutChar):0)

/*
 		#] OutPutMacros :
 		#[ Definitions of caps : - Only the necessary ones

	The ones with the minus sign in their description are already
	being used.

*/

BYTE *AE;			/* - end alternate character set */
BYTE *AL;			/* - insert line */
BOOLEAN AM;			/* - automatic margins (wrap around) */
BYTE *AS;			/* - Start alternate character set */
BYTE *BC;			/*   backspace (if not BS) */
BOOLEAN BM;			/*   blocking mode: sent complete strings. no economy. */
BYTE *CE;			/* - clear to end of line */
BYTE *CH;			/* + move cursor horizontally */
BYTE *CL;			/* - clear screen */
BYTE *CM;			/* - cursor motion */
WORD CO;			/* - number of columns */
BYTE *CS;			/* - set scroll region */
BYTE *CV;			/* + move cursor vertically */
BYTE *DC;			/* . delete character */
BYTE *DL;			/* - delete line */
BYTE *DM;			/* . enter delete mode */
BYTE *DO;			/* - cursor down */
BYTE *ED;			/* . end delete mode */
BYTE *EI;			/* . end insert mode */
BOOLEAN HC;			/* - hardcopy terminal */
BYTE *IC;			/* . insert character */
BYTE *IM;			/* . enter insert mode */
BOOLEAN IN;			/* . Insertmode distinguishes NULL on display */
BYTE *IP;			/* . post-insert padding */
BYTE *LE;			/* - Move cursor left */
WORD LI;			/* - number of rows */
BYTE *YE;			/* - Unset inverse mode */
BYTE *YS;			/* - Set inverse mode */
BYTE PC, *PCstr;	/* - pad character and string */
BYTE *RI;			/* - move cursor right */
BYTE *ZS;			/* - Show cursor */
BYTE *SE;			/* - end standout mode */
BYTE *SF;			/* - scroll forwards */
BYTE *SO;			/* - enter standout mode */
BYTE *SR;			/* - scroll reverse */
BYTE *TE;			/* - end cursor addressing mode */
BYTE *TI;			/* - enter cursor addressing mode */
BYTE *UP;			/* - cursor up */
BYTE *VE;			/* - end visual mode */
BYTE *VS;			/* - enter visual mode */
BYTE *WE;			/* - end wrap mode */
BYTE *WS;			/* - enter wrap mode */
BYTE *ZE;			/* - Cursor off */

BYTE *SETINV;
BYTE *UNSETINV;
BYTE *SETWRAP;
BYTE *UNSETWRAP;
/*
	This is plumbing for VT100. Is not in some termcap files.
*/
BYTE SetInv[] = "\033[?5h";
BYTE UnsetInv[] = "\033[?5l";
BYTE SetWrap[] = "\033[?7h";
BYTE UnsetWrap[] = "\033[?7l";
/*
	And some plumbing for VT320 special keys
*/
BYTE DoShDown[] = "\033[6~";
BYTE DoShUp[] = "\033[5~";
BYTE DoInKey[] = "\033[2~";
BYTE DoDelLin[] = "\033[3~";

BYTE HideCursor[10];
BYTE ShowCursor[10];

BYTE KeyUp1[] = "\033[A";
BYTE KeyDown1[] = "\033[B";
BYTE KeyLeft1[] = "\033[D";
BYTE KeyRight1[] = "\033[C";
BYTE KeyUp2[] = "\033[OA";
BYTE KeyDown2[] = "\033[OB";
BYTE KeyLeft2[] = "\033[OD";
BYTE KeyRight2[] = "\033[OC";

BYTE AltKey[] = "\033[211z";
BYTE CtrlKey[] = "\033[212z";
BYTE NfunKey[] = "\033[213z";
BYTE SfunKey[] = "\033[254z";

/*
 		#] Definitions of caps :
  	#] Definitions :
  	#[ Abort routines :
 		#[ fatal :			VOID fatal(msg)
*/

VOID
fatal(BYTE *msg)
{
	PSend();
	fprintf(stderr, "ERROR: %s\n", msg);
	exit(1);
}

/*
 		#] fatal :
 		#[ syserr :			VOID syserr(msg)
*/

VOID
syserr(BYTE *msg)
{
	extern int errno, sys_nerr;
/*
	extern __const char *__const sys_errlist[];
*/
	PSend();
	fprintf(stderr, "ERROR: %s (%d",msg,errno);
	if ( errno > 0 && errno < sys_nerr)
		fprintf(stderr, "; %s)\n", sys_errlist[errno]);
	else
		fprintf(stderr, ")\n");
	exit(1);
}

/*
 		#] syserr :
 		#[ termfatal :		VOID termfatal(msg)
*/

VOID
termfatal(BYTE *msg)
{
	BYTE buf[100];
	sprintf(buf, "Terminal %s %s\n", getenv("TERM"), msg);
	fatal(buf);
}

/*
 		#] termfatal :
  	#] Abort routines :
  	#[ Screen routines :
 		#[ StrCont :		WORD StrCont(s1,s2)

		Returns 1 when string 1 is contained in string 2

*/

WORD
StrCont(BYTE *s1,BYTE *s2)
{
	while ( *s1 ) {
		if ( *s1++ != *s2++ ) return(0);
	}
	return(1);
}

/*
 		#] StrCont :
 		#[ getcaps :		VOID getcaps()

		Gets the needed terminal capabilities.
		Note that some capabilities were initialized the hard
		way, because the termcaps at NIKHEF-H didn't have them
		built in yet. See the definition field of the termcap
		entries.

*/

VOID
getcaps(VOID)
{
	static BYTE capbuf[2048];
	static BYTE tcbuf[2048];
	BYTE *area;
	BYTE *strsun;
#ifdef TIOCGWINSZ
/*	Mechanism to pick up the real size of the screen. Used by Apollo. */
    struct winsize window;
#endif

/*
		Get the terminal type and read its termcap into tcbuf.
*/
	switch ( tgetent(tcbuf,getenv("TERM")) ) {
		case -1:
			fatal("can't open termcap file");
		case 0:
			fatal("terminal not defined");
		case 1:
			break;
		default:
			fatal("bad return from tgetent");
	}
/*
		Read out the various capabilities:

		We are still missing some for the cursor motion.
		They concern the motion inside the same row or the same column
*/
	area = capbuf;
	AE = tgetstr("ae",&area);
	AL = tgetstr("al",&area);
	AM = tgetflag("am");
	AS = tgetstr("as",&area);
/*	BC = tgetstr("bc",&area);	*/
/*	BM = tgetflag("bm");		*/
	CE = tgetstr("ce",&area);
	CH = tgetstr("ch",&area);
	CL = tgetstr("cl",&area);
	CM = tgetstr("cm",&area);
	CO = tgetnum("co");
	CS = tgetstr("cs",&area);
	CV = tgetstr("cv",&area);
	DC = tgetstr("dc",&area);
	DL = tgetstr("dl",&area);
	DM = tgetstr("dm",&area);
	DO = tgetstr("do",&area);
	ED = tgetstr("ed",&area);
	EI = tgetstr("ei",&area);
	HC = tgetflag("hc");
	IC = tgetstr("ic",&area);
	IM = tgetstr("im",&area);
	IN = tgetflag("in");
	IP = tgetstr("ip",&area);
	LE = tgetstr("le",&area);
	LI = tgetnum("li");
	YE = tgetstr("ye",&area);
	YS = tgetstr("ys",&area);
	PCstr = tgetstr("pc",&area);
	RI = tgetstr("nd",&area);
	ZS = tgetstr("zs",&area);
	SE = tgetstr("se",&area);
	SF = tgetstr("sf",&area);
	SO = tgetstr("so",&area);
	SR = tgetstr("sr",&area);
	TE = tgetstr("te",&area);
	TI = tgetstr("ti",&area);
	UP = tgetstr("up",&area);
	VE = tgetstr("ve",&area);
	VS = tgetstr("vs",&area);
	WE = tgetstr("we",&area);
	WS = tgetstr("ws",&area);
	ZE = tgetstr("ze",&area);

	strsun = getenv("TERM");
	if ( strsun && ( StrCmp("sun",strsun) == 0 ) ) {
		AM=TRUE;
	}

	if ( !YE || !YS ) { SETINV = SetInv; UNSETINV = UnsetInv; }
	else { SETINV = YS; UNSETINV = YE; }

	if ( !WE || !WS ) { SETWRAP = SetWrap; UNSETWRAP = UnsetWrap; }
	else { SETWRAP = WS; UNSETWRAP = WE; }

	if ( !ZS || !ZE ) {
		if ( StrCont("vt220",getenv("TERM")) ) {
			strcpy(HideCursor,"\033[?25l");
			strcpy(ShowCursor,"\033[?25h");
		}
		else if ( StrCont("vt1",getenv("TERM"))
		|| StrCont("uniterm",getenv("TERM")) ) {
			strcpy(HideCursor,"\033[1v");
			strcpy(ShowCursor,"\033[0v");
		}
		else {
			strcpy(HideCursor,"");
			strcpy(ShowCursor,"");
		}
		ZS = ShowCursor;
		ZE = HideCursor;

		ZS = 0;
		ZE = 0;
	}


	if ( PCstr != NULL ) { PC = PCstr[0]; }				/* Pad string */
	else { PC = '\0'; }
	if ( !CM ) {										/* Cursor motion */
		termfatal("cannot work for STedi");
	}

#ifdef TIOCGWINSZ
    /*
     * fetch real window size, if this is a terminal,
     * and the window size is set.
     */
    if (ioctl(input_fd, TIOCGWINSZ, &window) == 0) {
		if (window.ws_row > 0 && window.ws_col > 0) {
	    	LI = window.ws_row;
	   		CO = window.ws_col;
		}
    }
#else
	if ( StrCont("apollo",getenv("TERM")) ) {
/*
		Ask for the size of the screen, Apollo style.
*/

	}
#endif
	if ( LI == -1 || HC ) {
		termfatal("is not a CRT");
	}
	if ( CO > MAXCOLS ) CO = MAXCOLS;
	if ( AM ) WrapMode = 1;
	else WrapMode = 0;
}

/*
 		#] getcaps :
 		#[ PSbegin :		VOID PSbegin()
*/

VOID
PSbegin(VOID)
{
	if ( psbeginfirst == 0 ) return;
	input_fd = fileno(stdin);
	psbeginfirst = 0;
	SetMyBuf();
	setbuf(stdout,0); 			/* Needed for direct screens */
	setraw();
	getcaps();
	if ( NotNormal && SETINV ) putctl(SETINV);
	if ( UNSETWRAP ) {
/*
	Current fix for VT100 etc. Can only be used when the automatic
	margins can be turned off. This is a cludge till the proper
	entries in the termcap files have been made.
*/
		putctl(UNSETWRAP);
		AM = 0;
		WrapMode = 0;
	}
	if ( getenv("STUPIDMARGINS") ) { CO--; AM = 0; }

	putctl(TI);
	putctl(VS);
	PSshowcur(FALSE);

	memsize = LI * CO;
	if ( ( LogMem  = (CELL)Malloc((LONG)(memsize*sizeof(CELL))) ) == 0
	  || ( PhysMem = (CELL)Malloc((LONG)(memsize*sizeof(CELL))) ) == 0 ) {
		fatal("No memory");
	}

	DefineWindow(&Screen,1,LI,1,CO,A_NORMAL,MASKS);
	DefineWindow(TextScreen,1,LI-1,1,CO,A_NORMAL,MASKS);
	DefineWindow(&Mes1Screen,LI,LI,1,5,A_INVERSE,MASKS);
	DefineWindow(&Mes2Screen,LI,LI,6,CO-21,A_INVERSE,MASKS);
	DefineWindow(&StatScreen,LI,LI,CO-20,CO,A_INVERSE,MASKS);

	PSclear(&Screen);
	if ( CS ) {
		ScrollTop = Screen.r1-1;
		ScrollBot = Screen.r2-2;
		putctl(tgoto(CS,ScrollBot,ScrollTop));
	}
	KeyIni();
}

/*
 		#] PSbegin :
 		#[ PSend :			VOID PSend()
*/

VOID
PSend(VOID)
{
	if ( psbeginfirst == 1 ) return;
	SetColor(A_NORMAL);
	PSshowcur(TRUE);
	if ( SETWRAP ) { putctl(SETWRAP); AM = 1; }
	if ( CS ) putctl(tgoto(CS,Screen.r2-1,Screen.r1-1));
	if ( UNSETINV ) putctl(UNSETINV);
	putctl(CL);
	putctl(VE);
	putctl(TE);
	KeyEnd();
	unsetraw();
	MyFlush();
	UndoMyBuf();
	if ( LogMem ) { Mfree(LogMem); LogMem = 0; }
	if ( PhysMem ) { Mfree(PhysMem); PhysMem = 0; }
	psbeginfirst = 1;
}

/*
 		#] PSend :
 		#[ PSwrite :		VOID PSwrite(scrn,row,col,ncols,str,mode)

		Writes the string str into the logical screen.
		It writes ncols columns.
		scrn is the representation of the window into which the
		string is written. It is written at row,col in this window.
		Note that row and col count from 1.

		mode can have the value A_CLEAR when a clear to end of line
		is desired after the characters have been written.
		other vaules are A_ALL or A_NORMAL.
*/

VOID
PSwrite(SCREEN *scrn,WORD row,WORD col,WORD ncols,BYTE *str,WORD mode)
{
	WORD leftover;
	CELL s;
/*
	Clipping.
	1:	Rows.
	2:	Columns left
	3:	Columns right
*/
	if ( row > scrn->hi || row < 1 ) return;
	if ( col < 1 ) {
		str += 1 - col;
		ncols -= 1 - col;
		col = 1;
	}
	if ( col > scrn->wi ) return;
	if ( ncols < 0 ) ncols = 0;
	leftover = scrn->wi + 1 - col - ncols;
	if ( leftover < 0 ) {			/* How many blanks at end for clear? */
		ncols = scrn->wi + 1 - col;
		leftover = 0;
	}
/*
	Now the drawing. First get the memory address in the logical screen:
*/
	s = LogMem + (scrn->ro + row - 1)*Screen.wi + (scrn->co + col - 1);
	while ( --ncols >= 0 ) {
		s->chr = *str++;
		s->att = scrn->att;
		s++;
	}
	if ( mode == A_CLEAR ) {
		while ( --leftover >= 0 ) {
			s->chr = ' ';
			s->att = scrn->att;
			s++;
		}
	}
}

/*
 		#] PSwrite :
 		#[ PSclreol :		VOID PSclreol(scrn,row,col)
*/

VOID
PSclreol(SCREEN *scrn,WORD row,WORD col)
{
	PSwrite(scrn,row,col,0,(BYTE *)0,A_CLEAR);
}

/*
 		#] PSclreol :
 		#[ PSsynch :		VOID PSsynch()

		Synchronization of the logical screen with the physical screen.
		Economization procedures are:
		Don't draw parts that are identical at the beginning of a line.
		Don't draw parts that are identical at the end of a line.
		Try to use CLRTOEOL to draw blanks at the end of a line.
		If a combination of the two end variations is better, use that.

		Special attention has to be given to the screen attributes
		(normal vs inverse) and the automatic margins (wrapping mode).
		It is preferred that the wrapping mode is off, because otherwise
		we cannot write at some positions.

		In the future we may still try to put inserts and deletes of
		characters in.
*/

VOID
PSsynch(VOID)
{
	WORD row,col,j,a,i,k,col2,m,aold;
	CELL s1,s2,t1,t2,u1,u2;
/*
	WORD oldrow = CurrentRow;
	WORD oldcol = CurrentCol;
*/
	for ( row = 1; row <= Screen.r2; row++ ) {
		t1 = LogMem + row * Screen.wi;
		t2 = PhysMem + row * Screen.wi;
		s1 = t1 - Screen.wi;
		s2 = t2 - Screen.wi;
		col = 1;
		j = Screen.wi;
		if ( WrapMode && row == Screen.r2 ) { j -= 2; t1 -= 2; t2 -= 2; }
/*
		strip same pieces from the start
*/
		while ( s1->chr == s2->chr && s1->att == s2->att && j > 0 )
			{ s1++; s2++; j--; col++; }
		t1--; t2--;
		a = t1->att;
		i = k = j;
/*
		strip same pieces from the end
*/
		u1 = t1; u2 = t2;
		while ( u1->chr == u2->chr && u1->att == u2->att && k > 0 )
			{ u1--; u2--; k--; }
/*
		strip spaces at the end and then same pieces at the end
*/
		while ( t1->chr == ' ' && j > 0 && t1->att == a ) { t1--; j--; t2--; }
		i = i-j;
		col2 = col+j; m = j;
		while ( t1->chr == t2->chr && t1->att == t2->att && j > 0 )
			{ t1--; t2--; j--; }
		m = m-j;
/*
		if k > j we use the model with the spaces at the end,
		otherwise we use the model in which we skip a part at the end.
*/
		if ( k > j ) {
			if ( j > 0 ) {
				SetCursor(row,col);
				while ( j-- > 0 ) {
					SetColor(s1->att);
					putchc(s1->chr);
					s2->chr = s1->chr;
					s2->att = s1->att;
					s1++; s2++;
					CurrentCol++;
				}
			}
			else if ( i > 0 && m <= 0 ) SetCursor(row,col);
			if ( i > 0 ) {
				if ( m > 0 ) {
					s2 += m;
					SetCursor(row,col2);
				}
				aold = 0;
				SetColor(a);
				while ( --i >= 0 ) {
					s2->chr = ' ';  s2->att = a;
					if ( a != A_NORMAL ) { putchc(' '); aold = 1; }
					s2++;
				}
				if ( !aold ) CLRTOEOL;
				else SetCursor(row,col2);
			}
		}
		else {
			if ( k > 0 ) {
				SetCursor(row,col);
				while ( --k >= 0 ) {
					SetColor(s1->att);
					putchc(s1->chr);
					s2->chr = s1->chr;
					s2->att = s1->att;
					s1++; s2++;
					CurrentCol++;
				}
			}
		}
		if ( WrapMode && CurrentCol > Screen.wi ) {
			CurrentCol = 1;
			CurrentRow++;
		}
	}

	if ( commandmode ) {
		PSsetcur(&Mes2Screen,1,3+(commandbuffer->column-commandbuffer->offset));
	}
	else {
		PSsetcur(TextScreen,cbuf->screenline,cbuf->screencolumn);
	}
/*
	SetCursor(oldrow,oldcol);
*/
	MyFlush();
}

/*
 		#] PSsynch :
 		#[ PSdraw :			VOID PSdraw()

		Forces a redraw of the screen.
		This is used when the screen has been messed up by
		some system message like the completely demented
		message `coffee time' etc.

*/

VOID
PSdraw(VOID)
{
	CELL s, t;
	WORD row,col,i,j,a;
	WORD oldrow = CurrentRow;
	WORD oldcol = CurrentCol;
	MessageForce();
	CurrentAtt = -1;
	SetColor(PhysMem->att);
	for ( row = 1; row <= Screen.r2; row++ ) {
		t = PhysMem + row*Screen.wi;
		s = t - Screen.wi;
		t--; a = t->att;
		j = Screen.wi;
		if ( WrapMode && row == Screen.r2 ) { j -= 2; t -= 2; }
		i = j;
		if ( a == A_NORMAL ) {
			while ( t->chr == ' ' && t->att == a && i > 0 ) { t--; i--; }
		}
		col = i; j = j - i;
		SetCursor(row,1);
		if ( i > 0 ) {
			while ( --i >= 0 ) {
				SetColor(s->att);
				if ( s->chr == 0 ) putchc(' ');
				else putchc(s->chr);
				s++;
				CurrentCol++;
			}
		}
		if ( j > 0 ) {
			SetColor(a);
			if ( row == Screen.c2 && WrapMode ) {
				while ( --j >= 0 ) putchc(' ');
			}
			else {
				CLRTOEOL;
			}
		}
		if ( WrapMode && CurrentCol > Screen.wi ) {
			CurrentCol = 1;
			CurrentRow++;
		}
	}
	SetCursor(oldrow,oldcol);
	MyFlush();
}
 
/*
 		#] PSdraw :
 		#[ PSclear :		VOID PSclear(scrn)
*/

VOID
PSclear(SCREEN *scrn)
{
	CELL s1,s2;
	WORD i;
	if ( scrn == &Screen ) {
		s1 = LogMem;
		s2 = PhysMem;
		i = memsize;
		while ( --i >= 0 ) {
			s1->chr = s2->chr = ' ';
			s1->att = s2->att = scrn->att;
			s1++; s2++;
		}
		SetColor(scrn->att);
		putctl(CL);
		CurrentRow = CurrentCol = 1;
	}
	else {
		for ( i = 1; i <= scrn->hi; i++ ) {
			PSwrite(scrn,i,1,0,(BYTE *)0,A_CLEAR);
		}
		PSsynch();
		SetCursor(scrn->r1,scrn->c1);
	}
}

/*
 		#] PSclear :
 		#[ PSsetcur :		VOID PSsetcur(scrn,row,col)

		Sets the cursor to a window position.
		rows and columns count from 1 to scrn->hi and scrn->wi resp.

*/

VOID
PSsetcur(SCREEN *scrn,WORD row,WORD col)
{
	if ( row <= 0 ) row = 1;
	else if ( row > scrn->hi ) row = scrn->hi;
	if ( col <= 0 ) col = 1;
	else if ( col > scrn->wi ) col = scrn->wi;
	if ( WrapMode && col == Screen.wi && row == Screen.hi ) col--;
	SetCursor(scrn->ro+row,scrn->co+col);
	MyFlush();
}

/*
 		#] PSsetcur :
 		#[ SetCursor :		VOID SetCursor(row,col)

		We need sequences for:
		goto column 1:	('\r')
		left arrow
		right arrow
		left n places
		right n places
		up arrow
		down arrow
		down one line ('\n')
		up n places
		down n places
		gotoXY (CM = Cursor motion)

*/

VOID
SetCursor(WORD row,WORD col)
{
	WORD difcol, difrow;
	if ( col > Screen.c2 ) col = Screen.c2;
	else if ( col < Screen.c1 ) col = Screen.c1;
	if ( row < Screen.r1 ) row = Screen.r1;
	else if ( row > Screen.r2) row = Screen.r2;
	if ( col == 1 && col != CurrentCol ) {
		putchs('\015'); CurrentCol = 1;
	}
	if ( row == CurrentRow ) {
		if ( col == CurrentCol ) return;
		difcol = col - CurrentCol;
		if ( difcol == 1 && RI ) {
			putctl(RI);
		}
		else if ( difcol == -1 && LE ) {
			putctl(LE);
		}
		else {
			if ( !CH ) goto General;
			tgoto(CH,col-1,col-1);
		}
		CurrentCol = col;
	}
	else if ( col == CurrentCol ) {
		difrow = row - CurrentRow;
		if ( difrow > 0 ) {
			if ( difrow == 1 && CurrentRow < (ScrollBot+1) && DO ) {
				putctl(DO);
			}
			else {
				if ( !CV ) goto General;
				tgoto(CV,row-1,row-1);
			}
		}
		else {
			if ( difrow == -1 && CurrentRow > (ScrollTop+1) && UP ) {
				putctl(UP);
			}
			else {
				if ( !CV ) goto General;
				tgoto(CV,row-1,row-1);
			}
		}
		CurrentRow = row;
	}
	else {
General:
		putctl(tgoto(CM,col-1,row-1));
		CurrentRow = row;
		CurrentCol = col;
	}
}

/*
 		#] SetCursor :
 		#[ PSscroll :		VOID PSscroll(scrn,num,row1,row2)

		Scrolls a section of the screen. We assume that this involves
		the full width of the screen, because terminals and termcaps
		have no provision for rectangular subparts of the screen.
		This could of course change in the future. A simple fixup
		of this routine will then allow the presence of the vertical
		split of the screen, which is now disabled.

		row1:	The top line of the region to be scrolled.
		row2:	The bottom line of the region to be scrolled (inclusive).
		num:	The number of lines to scroll inside this region.
				num>0	scroll down
				num<0	scroll up.

		Note that the lines are scrolled one by one to allow an easy
		clearing of the new lines. It is by no means guaranteed that
		the terminal will do this for us, although some will.

		We need either:
		CS		set scrolling region
		SR		scroll reverse
		SF		scroll forward
		or:
		DL		delete line
		AL		insert line

		When the terminal (or most likely the termcap entry) doesn't
		have a scrolling capability we will get the very distasteful
		dancing bottom line(s).
*/

VOID
PSscroll(SCREEN *scrn,WORD num,WORD row1,WORD row2)
{
	CELL s1, s2, t1, t2;
	WORD Top, Bot, i1, i2, offset, i;
	SetColor(scrn->att);
	if ( num > 0 ) {				/* Scroll region down */
		offset = ( scrn->r1 + row2 ) * Screen.wi;
		i1 = num * Screen.wi;
		s1 = LogMem + offset; s2 = PhysMem + offset;
		t1 = s1 - i1; t2 = s2 - i1;
		i2 = ( row2 - row1 + 2 ) * Screen.wi - i1;
		while ( i2-- > 0 ) {
			s1--; s2--; t1--; t2--;
			s1->chr = t1->chr; s2->chr = t2->chr;
			s1->att = t1->att; s2->att = t2->att;
		}
		while ( i1-- > 0 ) {
			s1--; s2--;
			s1->chr = s2->chr = ' ';
			s1->att = s2->att = scrn->att;
		}
		if ( CS ) {
			Top = row1+scrn->ro-1;
			Bot = row2+scrn->ro;
			if ( Top != ScrollTop || Bot != ScrollBot ) {
				putctl(tgoto(CS,Bot,Top));
				ScrollTop = Top; ScrollBot = Bot;
				CurrentCol = CurrentRow = 1;
			}
			PSsetcur(scrn,row1,1);
			while ( --num >= 0 ) {
				putctl(SR);
				CLRTOEOL;
			}
		}
		else {
/*			PSsetcur(scrn,row2-num+1,1); */
			PSsetcur(scrn,row2-num+2,1);
			i = num;
			while ( --i >= 0 ) putctl(DL);
			PSsetcur(scrn,row1,1);
			while ( --num >= 0 ) putctl(AL);
		}
	}
	else if ( num < 0 ) {	/* Scroll region up */
		num = -num;
		offset = ( scrn->ro + row1 - num - 1 ) * Screen.wi;
		if ( offset < 0 ) offset = 0;
		i1 = num * Screen.wi;
		s1 = LogMem + offset; s2 = PhysMem + offset;
		t1 = s1 + i1; t2 = s2 + i1;
		i2 = ( row2 - row1 + 2 ) * Screen.wi - i1;
		while ( i2-- > 0 ) {
			s1->chr = t1->chr; s2->chr = t2->chr;
			s1->att = t1->att; s2->att = t2->att;
			s1++; s2++; t1++; t2++;
		}
		while ( i1-- > 0 ) {
			s1->chr = s2->chr = ' ';
			s1->att = s2->att = scrn->att;
			s1++; s2++;
		}
		if ( CS ) {
			Top = row1+scrn->ro-2;
			Bot = row2+scrn->ro-1;
			if ( Top != ScrollTop || Bot != ScrollBot ) {
				putctl(tgoto(CS,Bot,Top));
				ScrollTop = Top; ScrollBot = Bot;
				CurrentCol = CurrentRow = 1;
			}
			PSsetcur(scrn,row2,1);
			while ( num-- > 0 ) {
				putctl(SF);
				CLRTOEOL;
			}
		}
		else {
/*
			PSsetcur(scrn,row1,1);
			i = num;
			while ( i-- > 0 ) putctl(DL);
			PSsetcur(scrn,row2-num+1,1);
			while ( i-- > 0 ) putctl(AL);
*/
			PSsetcur(scrn,row1-1,1);
			i = num;
			while ( i-- > 0 ) putctl(DL);
			PSsetcur(scrn,row2-num+1,1);
			i = num;
			while ( i-- > 0 ) putctl(AL);
		}
	}
}

/*
 		#] PSscroll :
 		#[ PSshowcur :		VOID PSshowcur(on)

		Note that this doesn't work on all terminals

*/

VOID
PSshowcur(WORD on)
{
	if ( on ) {
		if ( !VisCursor ) {
			if ( ZS ) putctl(ZS);
			VisCursor = 1;
		}
	}
	else {
		if ( VisCursor ) {
			if ( ZE ) putctl(ZE);
			VisCursor = 0;
		}
	}
	MyFlush();
}

/*
 		#] PSshowcur :
 		#[ PSbeep :			VOID PSbeep()
*/

VOID
PSbeep(VOID)
{
	putchs(BELL);
	MyFlush();
}

/*
 		#] PSbeep :
 		#[ PScolor :		VOID PScolor()

		Toggle the color of the screen.

*/

VOID
PScolor(VOID)
{
	if ( !NotNormal ) { if ( SETINV ) { putctl(SETINV); NotNormal = 1; } }
	else { if ( UNSETINV ) { putctl(UNSETINV); NotNormal = 0; } }
	MyFlush();
}

/*
 		#] PScolor :
 		#[ SetColor :		VOID SetColor(att)

		Sets the color attribute of the screen.
		Needs SE	end standout mode
		      SO	enter standout mode

*/

VOID
SetColor(WORD att)
{
	if ( att != CurrentAtt ) {
		if ( att == A_INVERSE ) { putctl(SO); }
		else { putctl(SE); }
		CurrentAtt = att;
	}
}

/*
 		#] SetColor :
 		#[ SetMyBuf :		VOID SetMyBuf()
*/
 
VOID
SetMyBuf(VOID)
{
	if ( ObStart == 0 ) {
		ObStart = (UBYTE *)Malloc(4112L);
	}
	ObPoin = ObStart;
	ObWrite = ObStart + 4096;
}

/*
 		#] SetMyBuf :
 		#[ UndoMyBuf :		VOID UndoMyBuf()
*/
 
VOID
UndoMyBuf(VOID)
{
	if ( ObStart ) { Mfree(ObStart); ObStart = 0; }
}

/*
 		#] UndoMyBuf :
 		#[ MyFlush :		VOID MyFlush()
*/
 
VOID
MyFlush(VOID)
{
	WORD towrite, written;
	if ( ObPoin > ObStart ) {
		towrite = ObPoin - ObStart;
		ObPoin = ObStart;
		while ( ( written = write(fileno(stdout),ObPoin,towrite) ) < towrite ) {
			if ( written > 0 ) {
				towrite -= written;
				ObPoin += written;
			}
		}
		ObPoin = ObStart;
	}
}

/*
 		#] MyFlush :
 		#[ OutChar :		int OutChar(c)
*/

int
OutChar(int c)
{
	if ( ( c == 0x1B && ObPoin > ( ObWrite - 20 ) ) || ObPoin >= ObWrite ) {
		MyFlush();
	}
	*ObPoin++ = c;
	return(0);
}

/*
 		#] OutChar :
 		#[ putchc :			VOID putchc(c)
*/

VOID
putchc(UBYTE c)
{
	if ( c < ' ' || c >= 0x80 ) {
		if ( c < ' ' ) c = '^';
		else if ( ( c & 0x80 ) != 0 ) c = '~';
	}
	OutChar(c);
}

/*
 		#] putchc :
 		#[ putchs :			VOID putchs(c)
*/

VOID
putchs(UBYTE c)
{
	OutChar(c);
}

/*
 		#] putchs :
  	#] Screen routines :
  	#[ Keyboard :
 		#[ Variables :
*/

#define MAXREDEFSPACE 512

LONG InputFlags = 0;
BYTE HardBuffer[12];
WORD InHardBuf = 0;
BYTE HardRedefs[MAXREDEFSPACE+8];
WORD NumHardRedefs = 0;
WORD EscapeFlag = 0;
LONG KeyCode[MAXREDEFSPACE/4];

#define ESCAPEKEY 0x1B

#define KEYFLAG1 0x01000000L
#define KEYFLAG2 0x02000000L
#define KEYFLAG3 0x04000000L
#define KEYFLAG4 0x08000000L

#define MAXPOLLBUFF 4096

UBYTE buffedc[MAXPOLLBUFF];
WORD first,last;
UBYTE isabrake;
UBYTE *DirKBshift;

static struct sigaction	old_tty_vecs, new_tty_vecs;
static struct sigaction	old_win_vecs, new_win_vecs;
static struct sigaction	old_tstp_vecs, new_tstp_vecs;
static struct sigaction	old_cont_vecs, new_cont_vecs;

/*
 		#] Variables :
 		#[ setraw :
*/

WORD firstraw = TRUE;
WORD firstKey = TRUE;

VOID setraw(VOID)
{
	struct termio tbuf;
	if ( !firstraw ) return;
	if ( ioctl(input_fd,TCGETA,&tbufsave) == -1 ) return;  /* May be pipe */
	tbuf = tbufsave;
#ifdef CDEBUG
fprintf(fdebug,"set ioctl\n");
fprintf(fdebug,"c_cflag = 0x%x, c_iflag = 0x%x, c_oflag = 0x%x, c_lflag = 0x%x
		,c_cc[4] = %d, c_cc[5] = %d\n"
			,tbuf.c_cflag,tbuf.c_iflag,tbuf.c_oflag,tbuf.c_lflag
			,tbuf.c_cc[4],tbuf.c_cc[5]);
#endif
	ospeed = tbuf.c_cflag & CBAUD;
	tbuf.c_iflag &= ~(INLCR | ICRNL | IUCLC | ISTRIP | IXON | BRKINT);
	tbuf.c_oflag &= ~OPOST;
	tbuf.c_lflag &= ~(ICANON | ISIG | ECHO);
	tbuf.c_cc[0] = tbuf.c_cc[1] = tbuf.c_cc[2] = tbuf.c_cc[3] = 0;
	tbuf.c_cc[4] = 1;	/* min */
	tbuf.c_cc[5] = 0;	/* time */
	if ( ioctl(input_fd,TCSETA, &tbuf) == -1 ) syserr("setraw - ioctl");
	firstraw = FALSE;
}

/*
 		#] setraw :
 		#[ unsetraw :
*/

VOID
unsetraw(VOID)
{
	if ( firstraw ) return;
#ifdef CDEBUG
fprintf(fdebug,"unset ioctl\n");
#endif
	(VOID)ioctl(input_fd,TCSETA,&tbufsave);
	firstraw = TRUE;
}

/*
 		#] unsetraw :
 		#[ setblock :
*/

VOID
setblock(int fd,BOOLEAN on)
{
	static int blockf, nonblockf;
	static BOOLEAN first = TRUE;

	int flags;
	static int ndelay = O_NONBLOCK;

	if ( first ) {
		first = FALSE;
		if ( (flags = fcntl(fd, F_GETFL)) == -1 ) syserr("fcntl");
		blockf = flags & ~ndelay;	/* make sure O_NONBLOCK is off */
		nonblockf = flags | ndelay;	/* make sure O_NONBLOCK is on  */
	}
	if ( fcntl(fd,F_SETFL, on ? blockf : nonblockf) == -1 )
		syserr("fcntl2");
}

/*
 		#] setblock :
 		#[ readcond :
*/

int
readcond(int fd,char *buf,unsigned int nbytes,BOOLEAN block)
{
	WORD ret;
	keywaitflag = 1;
	setblock(fd,block);
	ret = read(fd,buf,nbytes);
	setblock(fd,!block);
	keywaitflag = 0;
	return(ret);
}

/*
 		#] readcond :
 		#[ KeyPoll :

		Check for a character from the keyboard.
		The value of par determines the time needed for a timeout.
		The timeout on a sequence that is going but not complete
		should be longer than the timeout on a fresh sequence.

WORD	isachar = 0;
char	buffedc[MAXPOLLBUF];

WORD
KeyPoll(WORD par)
{
	extern int errno;
	BYTE newchar;
	LONG ii;
	if ( isachar ) return(TRUE);
Again:
	switch (readcond(input_fd, &newchar, 1, par)) {
		case -1:
			if ( errno == EWOULDBLOCK ) {
				if ( isachar ) return(TRUE);
				return(FALSE);
			}
			syserr("read");
			return(FALSE);
		case 0:
			if ( isachar ) return(TRUE);
			return(FALSE);
		default:
			buffedc[isachar++] = newchar;
			if ( newchar == ESCAPEKEY && isachar < (MAXPOLLBUF-1) ) {
				par = 0;
				goto Again;
			}
			return(TRUE);
	}
}

*/

WORD
KeyPoll(WORD par)
{
	if ( last != first ) return(TRUE);
	else return(FALSE);
}

/*
 		#] KeyPoll :
 		#[ KeyGet :

		Wait for a character from the console.
		After the character has arrived: wait for a specific period
		to see whether it is followed by another character that
		may belong to an escape sequence.
		It is not yet clear what the best strategy is.

LONG
KeyGet(VOID)
{
	LONG ii;
	BYTE newchar;
	while ( !isachar ) KeyPoll(1);
	newchar = buffedc[0];
	isachar--;
	for ( ii = 0; ii < isachar; ii++ ) buffedc[ii] = buffedc[ii+1];
	return((LONG)(newchar&0xFF));
}
*/

LONG
KeyGet(VOID)
{
	WORD newchar;
	keywaitflag = 1;
	while ( last == first ) {
		sigpause(0);
	}
	keywaitflag = 0;
	newchar = buffedc[first];
	first = (first+1)%MAXPOLLBUFF;
	return((LONG)(newchar & 0xFF));
}

/*
 		#] KeyGet :
 		#[ KeyChar :

		Compares the characters coming through with the hard redefinitions
		coming either from the termcap or the definitions that are
		terminal dependent, but available on a separate file.

*/

LONG
KeyChar(VOID)
{
	BYTE *s, *t, c, cc;
	WORD i = 0, j = 0;
	BYTE *top;
	LONG k;
	if ( InHardBuf == 0 ) {
		HardBuffer[0] = c = KeyGet();
		InHardBuf++;
	}
	else c = HardBuffer[0];
/*
	if ( c == ESCAPEKEY && !isachar ) {
		k = ESCAPE; goto KeyAll;
	}
*/
	s = HardRedefs;
	top = HardRedefs + NumHardRedefs;
	cc = c;
	while ( s < top ) {
		t = s + *s;
		s++;
		if ( *s == c && s < t ) {		/* Partial sequence */
			do {
				s++;
				if ( s == t ) {
					k = KeyCode[j];
					goto KeyAll;
				}
				i++;
				if ( InHardBuf == i ) {
/*
----->				There skould be a rather liberal timeout here.
*/
					HardBuffer[i] = c = KeyGet();
					InHardBuf++;
				}
				else c = HardBuffer[i];
			} while ( *s == c );
			i = 0;
			c = HardBuffer[0];
		}
		s = t;
		j++;
	}

	k = c & 0xFFL;
	if ( ( k & 0x80 ) != 0 ) {
		k = (k&0x7F) | KEYFLAG1;
	} else
	switch ( (WORD)k ) {
		case 0x1B:
			k = ESCAPE;
			break;
		case 0x08:
			k = BACKSPACE;
			break;
		case 0x7F:
			k = DELETE;
			break;
		case 0x0A:
			k = LINEFEED;
			break;
		case 0x0D:
			k = CRETURN;
			break;
		case 0x02:
			k = SH_UP;
			break;
		case 0x06:
			k = SH_DOWN;
			break;
		case 0x0B:
			k = SH_LEFT;
			break;
		case 0x0C:
			k = SH_RIGHT;
			break;
		case 0x1F:
/*
			There is a problem here. We would like to have
			ctrl-something = delete line.
			We also like to keep the date facility.
			Keep the date for now.
*/
			k = CT_MIN;
			break;
		case 9:
			break;
		default:
			if ( !k ) k = CTRL_2;
			else if ( k <= 0x1D ) k += 0x240;
			else if ( k <= 0x1F ) k += 0x220;
			break;
	}
KeyAll:
	for ( j = i + 1; j < InHardBuf; j++ ) HardBuffer[j-i-1] = HardBuffer[j];
	InHardBuf -= i + 1;
	return(k);
}

/*
 		#] KeyChar :
 		#[ KeyRead :

		Gets a complete key, including all flags.
		If the Key is a flag key we turn on the corresponding flag
		and ask for the next key.
		If the key is not a flag key we combine it with the input flags
		and clear those flags.
		Finally we combine with the MacroFlags.

*/

LONG
KeyRead(VOID)
{
	LONG k;
	LONG InputF;
	for (;;) {
		k = KeyChar();
		if ( k & 0xFF000000L ) InputFlags |= k;
		else {
			if ( InputFlags ) {
/*
				Temporary solution for the lack of insert and other
				keys on most terminals. The best solution would be
				an appendix to the keyredefinitions.
*/
				InputF = InputFlags;
				InputFlags = 0;
				if ( InputF == KEYFLAG1 ) {
					if ( k == BACKSPACE )		/* Alt-backspace */
						return ( UNDO | KEYFLAG1 );
					if ( k == '?' )
						return ( HELP | KEYFLAG4 );
				}
				else if ( InputF == KEYFLAG2 ) {
					if ( k == DELETE )
						return ( CT_DELETE | KEYFLAG2 );
				}
				else if ( InputF == KEYFLAG3 ) {
					if ( k == '.' )
						return ( HOME | KEYFLAG3 );
				}
				else if ( InputF == KEYFLAG4 ) {
					if ( k == '-' )
						return ( INSERT | KEYFLAG4 );
					if ( k == '+' || k == ',' )
						return ( SH_INSERT | KEYFLAG4 );
					if ( k == '.' )
						return ( SH_HOME | KEYFLAG4 );
				}
/*
				Now the 'normal work'
				First remap the PF3 and PF4 key combinations.
*/
				if ( ( InputF & (KEYFLAG3|KEYFLAG4) ) == InputF ) {
					k -= '0';
					if ( k == 0 ) k = 10;
				}
				else if ( k >= 'a' && k <= 'z' ) k -= 'a' -'A';
				k |= ( InputF >> 16 );
				k |= InputF;
				InputF = 0;
			}
			return( k | MacroFlags );
		}
	}
}

/*
 		#] KeyRead :
 		#[ KeyStart :
*/

#define ADDKEY(nn,x) if ( ( cap = tgetstr(nn,&area) ) != 0 ){ t = s++; i = 0;\
 while( ( *s = *cap++ ) != 0 ){ i++; s++; } *t = i+1; *to++ = x; if ( s > top ) return(-1);}

#define ADDKEYCODE(nn,x) { cap = nn; t = s++; i = 0; while( ( *s\
 = *cap++ ) != 0 ){ i++; s++; } *t = i+1; *to++ = x; if ( s > top ) return(-1);}

WORD
KeyStart(VOID)
{
	BYTE capbuf[512];
	BYTE *s, *t, *top;
	WORD i, n;
	LONG *to, code = 0;
	BYTE *area, *cap, *kocap;
	area = capbuf;
	s = t = HardRedefs;
	to = KeyCode;
	top = s + MAXREDEFSPACE;
	ADDKEY("ku",C_UP)
	ADDKEY("kd",C_DOWN)
	ADDKEY("kl",C_LEFT)
	ADDKEY("kr",C_RIGHT)
	ADDKEY("kh",HOME)
#ifdef NEXT
	ADDKEY("kb",DELETE)
#else
	ADDKEY("kb",BACKSPACE)
#endif
/*
	The next code is far from perfect. In the corresponding l code
	there is the potential name of a key
	This should be used. It makes things complicated though as it
	leaves it up in the air which keys generate the keyflags.
*/
	ADDKEY("k1",KEYFLAG1)
	ADDKEY("k2",KEYFLAG2)
	ADDKEY("k3",KEYFLAG3)
	ADDKEY("k4",KEYFLAG4)
	ADDKEYCODE(AltKey,KEYFLAG1)
	ADDKEYCODE(CtrlKey,KEYFLAG2)
	ADDKEYCODE(NfunKey,KEYFLAG3)
	ADDKEYCODE(SfunKey,KEYFLAG4)
	ADDKEY("k5",HELP)
	ADDKEY("k6",UNDO)
	ADDKEYCODE("\033[234z",HELP)
	ADDKEYCODE("\033[235z",UNDO)
	ADDKEY("ec",ESCAPE)
	ADDKEY("bc",BACKSPACE)
	ADDKEYCODE("\033[214z",HOME);
	ADDKEYCODE("\033[220z",SH_HOME);
	ADDKEYCODE("\033[216z",SH_UP);
	ADDKEYCODE("\033[222z",SH_DOWN);
	ADDKEYCODE("\033[247z",INSERT);
	ADDKEYCODE("\033[249z",CT_DELETE);
	ADDKEYCODE("\033OP",KEYFLAG1);
	ADDKEYCODE("\033OQ",KEYFLAG2);
	ADDKEYCODE("\033OR",KEYFLAG3);
	ADDKEYCODE("\033OS",KEYFLAG4);
	if ( StrCont("xterm",getenv("TERM")) 
	|| StrCont("kterm",getenv("TERM")) ) {
		ADDKEYCODE("\033[11~",KEYFLAG1);
		ADDKEYCODE("\033[12~",KEYFLAG2);
		ADDKEYCODE("\033[13~",KEYFLAG3);
		ADDKEYCODE("\033[14~",KEYFLAG4);
		ADDKEYCODE("\033[15~",HELP);
		ADDKEYCODE("\033[17~",UNDO);
		ADDKEYCODE("\033[2~",INSERT);
		ADDKEYCODE("\033[5~",SH_UP);
		ADDKEYCODE("\033[6~",SH_DOWN);
	}
/*
	Code for the NeXT
*/
#ifdef NEXT
	ADDKEYCODE("\033\033",ESCAPE);
	ADDKEYCODE("\033a",0x0141);
	ADDKEYCODE("\033b",0x0142);
	ADDKEYCODE("\033c",0x0404);
	ADDKEYCODE("\033C",0x0804);
	ADDKEYCODE("\033d",0x0144);
	ADDKEYCODE("\033e",0x0145);
	ADDKEYCODE("\033f",0x0146);
	ADDKEYCODE("\033g",0x0147);
	ADDKEYCODE("\033h",0x0148);
	ADDKEYCODE("\033i",0x0149);
	ADDKEYCODE("\033j",0x014a);
	ADDKEYCODE("\033k",0x014b);
	ADDKEYCODE("\033l",0x014c);
	ADDKEYCODE("\033m",0x0401);
	ADDKEYCODE("\033n",0x014e);
	ADDKEYCODE("\033o",0x014f);
	ADDKEYCODE("\033p",0x0150);
	ADDKEYCODE("\033q",0x0151);
	ADDKEYCODE("\033r",0x0152);
	ADDKEYCODE("\033s",0x0153);
	ADDKEYCODE("\033t",0x0154);
	ADDKEYCODE("\033u",0x0155);
	ADDKEYCODE("\033v",0x0405);
	ADDKEYCODE("\033V",0x0805);
	ADDKEYCODE("\033w",0x0157);
	ADDKEYCODE("\033x",0x0403);
	ADDKEYCODE("\033X",0x0803);
	ADDKEYCODE("\033y",0x0159);
	ADDKEYCODE("\033z",0x015a);
	ADDKEYCODE("\0330",0x0130);
	ADDKEYCODE("\0331",0x0131);
	ADDKEYCODE("\0332",0x0132);
	ADDKEYCODE("\0333",0x0133);
	ADDKEYCODE("\0334",0x0134);
	ADDKEYCODE("\0335",0x0135);
	ADDKEYCODE("\0336",0x0136);
	ADDKEYCODE("\0337",0x0137);
	ADDKEYCODE("\0338",0x0138);
	ADDKEYCODE("\0339",0x0139);
	ADDKEYCODE("\033!",0x0231);
	ADDKEYCODE("\033@",0x0232);
	ADDKEYCODE("\033#",0x0233);
	ADDKEYCODE("\033$",0x0234);
	ADDKEYCODE("\033%",0x0235);
	ADDKEYCODE("\033^",0x0236);
	ADDKEYCODE("\033&",0x0237);
	ADDKEYCODE("\033*",0x0238);
	ADDKEYCODE("\033(",0x0239);
	ADDKEYCODE("\033)",0x0230);
	ADDKEYCODE("\033=",0x013D);
	ADDKEYCODE("\033\t",0x0109);
	ADDKEYCODE("\033\177",CT_DELETE);
	ADDKEYCODE("\177",BACKSPACE);
	ADDKEYCODE("\010",DELETE);
	ADDKEYCODE("\033?",0x101D);
#else
	ADDKEYCODE("\033[11~",FUN1);
	ADDKEYCODE("\033[12~",FUN2);
	ADDKEYCODE("\033[13~",FUN3);
	ADDKEYCODE("\033[14~",FUN4);
	ADDKEYCODE("\033[15~",FUN5);
	ADDKEYCODE("\033[17~",FUN6);
	ADDKEYCODE("\033[18~",FUN7);
	ADDKEYCODE("\033[19~",FUN8);
	ADDKEYCODE("\033[20~",FUN9);
	ADDKEYCODE("\033[21~",FUN10);
	ADDKEYCODE("\033[23~",KEYFLAG1);
	ADDKEYCODE("\033[24~",KEYFLAG2);
	ADDKEYCODE("\033[26~",UNDO);
	ADDKEYCODE("\033[28~",HELP);
	ADDKEYCODE("\033[34~",FUN3);
	ADDKEYCODE("\033[29~",FUN4);
	ADDKEYCODE("\033[32~",FUN5);
	ADDKEYCODE("\033[28~",HOME);
#endif
/*
	A dec key to keep Bert happy
*/
	ADDKEYCODE("\033[~4",ESCAPE);
	ADDKEYCODE("\033[4~",ESCAPE);
/*
	Next are some typical vt200/300 keys to make life easy
	There is no room in the termcaps for them
*/
	ADDKEYCODE(DoShUp,SH_UP)
	ADDKEYCODE(DoShDown,SH_DOWN)
	ADDKEYCODE(DoInKey,INSERT)
	ADDKEYCODE(DoDelLin,SH_UP)  /* PC-compatible to avoid horrible mistakes */
	ADDKEYCODE(KeyUp1,C_UP);
	ADDKEYCODE(KeyUp2,C_UP);
	ADDKEYCODE(KeyDown1,C_DOWN);
	ADDKEYCODE(KeyDown2,C_DOWN);
	ADDKEYCODE(KeyLeft1,C_LEFT);
	ADDKEYCODE(KeyLeft2,C_LEFT);
	ADDKEYCODE(KeyRight1,C_RIGHT);
	ADDKEYCODE(KeyRight2,C_RIGHT);


	if ( ( n = tgetnum("kn") ) > 0 ) {
		if ( ( kocap = tgetstr("ko", &area ) ) != 0 ) {
			while ( n-- > 0 ) {
				kocap[2] = 0;
/*
				Now figure out what capability this is.
				Put the corresponding code in `code'
				If capability not found, skip the following two lines.
*/
				ADDKEY(kocap,code);
				kocap += 3;
			}
		}
	}

/*
	Now read a potential additional file for this terminal

	This code has been skipped currently.

	The basic idea is:

	Get a terminaltype from getenv("TERM") which resides already in
	tcbuf. Take the name and append the extension .keys
	Look for this file in the directory /etc/termcap
	If not found try for the same file in the current directory
	If not found skip the code. Otherwise:
	Open the file and expect lines:
	Code=key1,key2,....,keyn;
	Everything after the ; is commentary.
	Each line that starts with a valid 8 digit hex number is a definition,
	all other lines are commentary. The keys are two digit hex numbers.

	There must be a list available of all codes accepted by the editor.

*/

	NumHardRedefs = s - HardRedefs;
	return(0);
}

/*
 		#] KeyStart :
 		#[ KeyFlush :

		A truncated version of KeyChar.
		If a complete character is present it is deleted.
		For safetysake it stops at the FLAG keys.
		The essence is to never delete a potentially partial sequence.


VOID
KeyFlush(VOID)
{
	BYTE *s, *t, c, cc;
	WORD i = 0, j = 0;
	BYTE *top;
	LONG k;
	for (;;) {
		if ( InHardBuf == 0 ) {
			if ( KeyPoll(0) == 0 ) return;
			HardBuffer[0] = c = KeyGet();
			InHardBuf++;
		}
		else c = HardBuffer[0];

		s = HardRedefs;
		top = HardRedefs + NumHardRedefs;
		cc = c;
		while ( s < top ) {
			t = s + *s;
			s++;
			if ( *s == c && s < t ) {
				do {
					s++;
					if ( s == t ) {
						k = KeyCode[j];
						goto KeyAll;
					}
					i++;
					if ( InHardBuf == i ) {
						if ( KeyPoll(0) == 0 ) return;
						HardBuffer[i] = c = KeyGet();
						InHardBuf++;
					}
					else c = HardBuffer[i];
				} while ( *s == c );
				i = 0;
				c = HardBuffer[0];
			}
			s = t;
			j++;
		}
KeyAll:
		if ( ( k & 0x0F000000L ) == k ) return;
		for ( j = i + 1; j < InHardBuf; j++ ) HardBuffer[j-i-1] = HardBuffer[j];
		InHardBuf -= i + 1;
	}
}
*/

VOID
KeyFlush(VOID)
{
	while ( KeyPoll(0) ) KeyChar();
	isabrake = 0;
	*DirKBshift = 0;
}

/*
 		#] KeyFlush :
 		#[ KeyInrupt :

		Interrupt routine for the keyboard.
		Characters are stored in the buffer `buffedc'.
		If there is a control-C the brake flags are set.
*/

void
KeyInrupt(int par)
{
	UBYTE c;
#ifdef CDEBUG
	fprintf(fdebug,"KeyInrupt(%d)\n",par);
#endif
	while ( read ( input_fd, &c, 1 ) == 1 ) {
#ifdef CDEBUG
		fprintf(fdebug,"0x%x\n",c);
#endif
		if ( c == 0x03 ) {
			isabrake = BREAKCODE;
		}
		else if ( c == 0x1A ) {
			kill(getpid(),SIGTSTP);
		}
		else if ( ( last + 1 ) % MAXPOLLBUFF != first ) {
			buffedc[last] = c;
			last = (last+1)%MAXPOLLBUFF;
		}
	}
#ifdef CDEBUG
	fprintf(fdebug,"0x----\n");
#endif
}

/*
 		#] KeyInrupt :
 		#[ Windows :
*/

VOID
WinInrupt(int par)
{
#ifdef CDEBUG
	fprintf(fdebug,"WinInrupt(%d)\n",par);
#endif
	if ( keywaitflag ) {
		PSend();
		PSbegin();
		ForceDisplay();
		MessageForce();
		PSsynch();
/*		PSdraw();  */
	}
	else winupflag = 1;
}

/*
 		#] Windows :
 		#[ TstpInrupt :
*/

VOID
TstpInrupt(int par)
{
#ifdef CDEBUG
	fprintf(fdebug,"TstpInrupt(%d)\n",par);
#endif
	if ( keywaitflag ) {
		PSend();
		old_tstp_vecs.sa_handler = SIG_DFL;
		sigaction ( SIGTSTP, &old_tstp_vecs, (struct sigaction *)0 );	

		new_cont_vecs.sa_handler = (void (*)(int))ContInrupt;
		sigemptyset(&(new_cont_vecs.sa_mask));
		new_cont_vecs.sa_flags = 0;
		sigaction ( SIGCONT, &new_cont_vecs, &old_cont_vecs );	

        kill(getpid(),SIGTSTP);
	}
}

/*
 		#] TstpInrupt :
 		#[ ContInrupt :
*/

VOID
ContInrupt(int par)
{
#ifdef CDEBUG
	fprintf(fdebug,"ContInrupt(%d)\n",par);
#endif
	old_cont_vecs.sa_handler = SIG_IGN;
	sigaction ( SIGCONT, &old_cont_vecs, (struct sigaction *)0);
 
	PSbegin();
	ForceDisplay();
	MessageForce();
	PSsynch();
}

/*
 		#] ContInrupt :
 		#[ StopHandler :

		Routine based on remarks from Chris Torek on usenet.

		Is still somewhat untested and didn't work the first
		time around at NIKHEF.
*/

VOID
StopHandler(VOID)
{
	int omask;
	PSend();

	omask = sigblock(0);
	kill(getpid(),SIGTSTP);
	sigpause(omask & ~sigmask(SIGTSTP));
 
	PSbegin();
	ForceDisplay();
	MessageForce();
	PSsynch();
}

/*
 		#] StopHandler :
 		#[ KeyIni :

VOID
KeyIni(VOID)
{
	setraw();
}
*/

void
KeyIni(VOID)
{
	int result;
	setraw();
	if ( !firstKey ) return;
#ifdef CDEBUG
	fprintf(fdebug,"Starting handlers\n");
#endif
	firstKey = 0;

	new_tty_vecs.sa_handler = (void (*)(int))KeyInrupt;
	sigemptyset(&(new_tty_vecs.sa_mask));
	new_tty_vecs.sa_flags = 0;

	new_win_vecs.sa_handler = (void (*)(int))WinInrupt;
	sigemptyset(&(new_win_vecs.sa_mask));
	new_win_vecs.sa_flags = 0;

	new_tstp_vecs.sa_handler = (void (*)(int))TstpInrupt;
	sigemptyset(&(new_tstp_vecs.sa_mask));
	new_tstp_vecs.sa_flags = 0;

	if ( !isatty ( input_fd ) )  {
		fprintf ( stderr, "input is not a keyboard!\n" );
		PSend ();
		exit (1);
	}
/*
	if ( !isatty ( 1 ) )  {
		fprintf ( stderr, "output is not a terminal!\n" );
		PSend ();
		exit (1);
	}
*/
	/* set up signal handler */
	sigaction ( SIGTSTP, &new_tstp_vecs, &old_tstp_vecs );
	sigaction ( SIGWINCH, &new_win_vecs, &old_win_vecs );	
	result = sigaction ( SIGIO, &new_tty_vecs, &old_tty_vecs );	
	if ( result != 0 )  {
		perror ( "sigaction: Error installing TTY interrupt handler\r\n" );
		PSend();
		exit (1);
	}
	/* enable asynchronous I/O on control tty */
	if ( firstKeyIni ) {
		firstKeyIni = 0;
		fcntlflags = fcntl(input_fd,F_GETFL);
	}
	result = fcntl ( input_fd, F_SETFL, O_ASYNC | O_NONBLOCK );
	if ( result != 0 )  {
		perror ( "fcntl: Error installing interrupt handler\r\n" );
		PSend();
		exit (1);
	}
	DirKBshift = &isabrake;
	isabrake = 0;
	first = last;
#ifdef CDEBUG
	fprintf(fdebug,"Handlers started\n");
#endif
}

/*
 		#] KeyIni :
 		#[ KeyEnd :

VOID
KeyEnd(VOID)
{
	unsetraw();
}
*/

void
KeyEnd(VOID)
{
	int result;
	if ( firstKey ) return;
	firstKey = 1;
	if ( !firstKeyIni ) {
#ifdef CDEBUG
fprintf(fdebug,"unset fcntl\n");
#endif
		fcntl(input_fd,F_SETFL,fcntlflags);
		firstKeyIni = 1;
	}
#ifdef CDEBUG
fprintf(fdebug,"unset handlers\n");
#endif
	old_tstp_vecs.sa_handler = SIG_IGN;
	sigaction ( SIGTSTP, &old_tstp_vecs, (struct sigaction *)0 );
	old_tty_vecs.sa_handler = SIG_DFL;
	result = sigaction ( SIGIO, &old_tty_vecs, (struct sigaction *)0 );
#ifdef CDEBUG
fprintf(fdebug,"tty handler result = %d\n",result);
#endif
	old_win_vecs.sa_handler = SIG_IGN;
	sigaction ( SIGWINCH, &old_win_vecs, (struct sigaction *)0 );
	unsetraw();
}

/*
 		#] KeyEnd :
 		#[ GetKBshift :
*/

UBYTE *
GetKBshift(VOID)
{
	return(&isabrake);
}

/*
 		#] GetKBshift :
  	#] Keyboard :
  	#[ Execute external programs :
 		#[ GetTicks :		LONG GetTicks()

		Should return a number of milli seconds.
		Could go wrong every 4.10^6 seconds!
		Something must be done about that.
		Not yet implemented on UNIX.

*/

#include <sys/time.h>
#include <sys/resource.h>

LONG
GetTicks(VOID)
{
	struct rusage rusage;
	getrusage(0,&rusage);
	return(rusage.ru_utime.tv_sec*1000+rusage.ru_utime.tv_usec/1000);
}

/*
 		#] GetTicks :
 		#[ Wait :

		For systems with very fast screen updates.
		It should wait till GetTicks returns a number larger than x.

*/

VOID
MyWait(LONG x)
{
}

/*
 		#] Wait :
 		#[ Excute :			WORD Excute(ProgramString,TailString,end,time)
*/

WORD
Excute(BYTE *ProgramString,BYTE *TailString,BYTE *end,LONG *time)
{
	LONG tim;
	WORD i;
	BYTE *s, *t;
	BYTE *command = 0;
	WORD ret;
	i = 0;
	s = ProgramString;
	while ( *s++ ) i++;
	s = TailString + 1;
	while ( *s++ ) i++;
	i += 2;
	if ( ( command = (BYTE *)(MALLOC((LONG)i)) ) == 0 ) {
		*time = 0;
		return(-8);
	}
	t = command;
	s = ProgramString;
	while ( ( *t++ = *s++ ) != 0 ) {}
	t[-1] = ' ';
	s = TailString + 1;
	while ( ( *t++ = *s++ ) != 0 ) {}
	t[-1] = '\0';
/*	PSend();  */
	tim = GetTicks();
	ret = system(command);
	*time = GetTicks() - tim;
/*	PSbegin(); */
	if ( command ) { Mfree(command); command = 0; }
	return(ret);
}

/*
 		#] Excute :
  	#] Execute external programs :
  	#[ Printer :

		A direct line to the printer can be established via a pipe.
		Currently not yet implemented.

*/

WORD
PrintOne(WORD handle,BYTE c)
{
	return(0);
}

WORD
IsAvailable(WORD device)
{
	return(0);
}

/*
  	#] Printer :
  	#[ System utilities :
 		#[ Free diskspace :

		For small computers: check the disk directly.
		On UNIX systems it should check for the users quota.

*/

LONG
DiskFree(UBYTE *s,WORD n)
{
	MesOut("Use `!quota -v' command");
	return(-1L);
}

/*
 		#] Free diskspace :
 		#[ IsBrake :

		Little problem: Needed to indicate whether a brake is
		active. May need a delicate interplay with the keyboard
		routines. A buffer is needed to store input characters,
		so that we may keep reading and look for ^C or someting.
		Currently not activated. Don't make loops!

*/

WORD IsBrake(VOID)
{
	return(0);
}

/*
 		#] IsBrake :
 		#[ IsCaps :

		Routine should tell whether the Capslock is active.
		Not possible for terminals. Was needed for Atari-ST.

*/

WORD
IsCaps(VOID)
{
	return(0);
}

/*
 		#] IsCaps :
 		#[ TakeScreen :		UBYTE *TakeScreen(size)

		Routine looks for some extra memory when the malloc for
		the write buffer fails. It is supposed to see whether
		the screen memory can be used and if it can it should
		return its address. Not possible in UNIX!

*/

UBYTE *
TakeScreen(LONG size)
{
	return((UBYTE *)0);
}

/*
 		#] TakeScreen :
  	#] System utilities :
*/
