/*
  	#[ Include file :
*/
#include "stedi.h"
/*
  	#] Include file :
  	#[ Auxiliary paste routines :
 		#[ Concat1 :

		Puts together the columns 1-'to-1' of line1 and
		all of line2. Result comes after line 'After'.

*/

WORD
Concat1(LINE Line1,WORD to,LINE Line2,LINE After)
/*
 *	Puts together the columns 1,(to-1) from line 1 with the whole
 *	of line 2. This comes after line 3. Watch the virtual blanks.
 */
{
	if ( to == 1 ) return ( CopyLine(Line2,After) );
	return(Concat3(Line1,to,Line2,1,After));
}

/*
 		#] Concat1 :
 		#[ Concat2 :

		Puts together all of line1 and the tail of line2.
		The tail starts at column 'from'. The result comes
		after the line 'After'.

*/

WORD
Concat2(LINE Line1,LINE Line2,WORD from,LINE After)
{
	WORD to;
	if ( Line1->used < 0 ) {
		if ( ( Line2->used != LINEHEAD ) || ( from != 1 ) )
												return((WORD)(NoFold()));
		return ( CopyLine(Line1,After) );
	}
	if ( Line2->used < 0 ) {
		if ( ( Line1->used != LINEHEAD ) || ( from != 1 ) )
												return((WORD)(NoFold()));
		return ( CopyLine(Line2,After) );
	}
	if ( Line1->used == LINEHEAD ) to = 1;
	else to = FindColumn(Line1,Line1->used-LINEHEAD) + 1;
	if ( Concat3(Line1,to,Line2,from,After) ) return(-1);
	return(to);
}

/*
 		#] Concat2 :
 		#[ Concat3 :

		Puts together the columns 1 - 'to-1' of Line1 and 'from' - end
		of Line2. The new line comes after line 'After'.
		The cut routine has to make sure that there are no foldlines.

*/

WORD
Concat3(LINE Line1,WORD to,LINE Line2,WORD from,LINE After)
{
	WORD type,num1,num2,virt1;
	WORD i, num3;
	UBYTE *s, *t;
	LINE l;
	FINDPOS fp;
	type = FindChar(Line1,to,&fp);
	virt1 = fp.virt; num1 = fp.num;
	if ( ( type != AFTERVIRTUAL ) && ( type != NORMALEOL ) ) num1--;
	type = FindChar(Line2,from,&fp);
	num2 = fp.num;
	if ( type == AFTERVIRTUAL || type == NORMALEOL ) {
		num2 = 0;
	}
	else {
/*		We used to make an attempt at killing unnecessary blanks inside
		tabs. This had rather bad consequences if indentation was deeper  */
/*		if ( type != NORMALCHAR ) virt1 = 0;  */
		num2 = Line2->used - LINEHEAD + 1 - num2;
	}
	num3 = num1 + num2 + virt1;
/*
	For debugging
*/
#ifdef DEBUGON
	if ( num1 < 0 || num2 < 0 || virt1 < 0 ) {
		for ( i = 0; i < 200; i++ ) MesOut("Bug in Concat3");
	}
#endif
	if ( num3 > MAXINLINE ) {
		MesnOut(4);
		if ( MesAsk(98,INCOMMAND) <= 0 ) return(-1);
		/* "Truncate too long line? (Y/N) " */
		num3 = MAXINLINE;
	}
	if ( ( l = LineAlloc(num3,After) ) == (LINE)0 ) return(NOMEMORY);
	s = l->text;
	i = num1;
	t = Line1->text;
	while ( ( --i >= 0 ) && ( --num3 >= 0 ) ) *s++ = *t++;
	i = virt1;
	while ( ( --i >= 0 ) && ( --num3 >= 0 ) ) *s++ = ' ';
	t = fp.address;
	while ( --num3 >= 0 ) *s++ = *t++;
	return(0);
}

/*
 		#] Concat3 :
 		#[ Cpline :

		Copy a range from a line.

*/

WORD
Cpline(LINE Line,WORD from,WORD to,LINE After)
{
	WORD type1, type2;
	LINE l;
	UBYTE *s, *t;
	WORD i = 0, j = 0;
	FINDPOS fp, fpp;
	type1 = FindChar(Line,from,&fp);
	type2 = FindChar(Line,to,&fpp);
	switch ( type2 ) {
		case ONTAB :
		case INSIDETAB :
				i = fpp.num - fp.num;
				j = fpp.virt;
				if ( i == 0 ) j -= fp.virt;
				break;
		case NORMALCHAR :
				i = fpp.num - fp.num;
				j = 0;
				break;
		case AFTERVIRTUAL :
		case NORMALEOL :
				if ( ( type1 == AFTERVIRTUAL ) || ( type1 == NORMALEOL ) ) {
					i = 0;
					j = fpp.virt - fp.virt;
				}
				else {
					i = fpp.num + 1 - fp.num;
					j = fpp.virt;
				}
	}
	if ( ( l = LineAlloc(i+j,After) ) == (LINE)0 ) return(NOMEMORY);
	s = l->text;
	t = fp.address;
	while ( --i >= 0 ) *s++ = *t++;
	while ( --j >= 0 ) *s++ = ' ';
	return(0);
}

/*
 		#] Cpline :
 		#[ CopyToEnd :
*/

WORD
CopyToEnd(LINE l,WORD from,LINE After)
{
	WORD n;
	n = LENGTH(l);
	n = FindColumn(l,n) + 1;
	return(Cpline(l,from,n,After));
}

 		/*
 		#] CopyToEnd :
 		#[ Dsline :

		Delete a range from a line.

*/

LINE
Dsline(LINE Line,WORD from,WORD to)
{
	WORD type,num1,num2,virt1;
	WORD i, num3;
	UBYTE *s, *t;
	LINE l;
	FINDPOS fp, fpp;
	type = FindChar(Line,from,&fp);
	num1 = fp.num; virt1 = fp.virt;
	if ( ( type != AFTERVIRTUAL ) && ( type != NORMALEOL ) ) num1--;
	type = FindChar(Line,to,&fpp);
	num2 = fpp.num;
	switch ( type ) {
		case ONTAB :
		case INSIDETAB :
				virt1 = 0;
		case NORMALCHAR :
				num2 = Line->used - LINEHEAD + 1 - num2;
				break;
		case AFTERVIRTUAL :
		case NORMALEOL :
				num2 = 0;
				break;
	}
	if ( num2 == 0 ) virt1 = 0;
	num3 = num1 + num2 + virt1;
	if ( num3 > MAXINLINE ) {
		virt1 -= ( num3 - MAXINLINE );
		num3 = MAXINLINE;
	}
	if ( ( num3 + LINEHEAD ) > Line->totlen ) {
		if ( ( l = LineAlloc(num3,Line) ) == (LINE)0 ) {
			virt1 = 0;
			num3 = num1 + num2;
		}
		else {
			s = l->text;
			i = num1;
			t = Line->text;
			while ( ( --i >= 0 ) && ( --num3 >= 0 ) ) *s++ = *t++;
			i = virt1;
			while ( ( --i >= 0 ) && ( --num3 >= 0 ) ) *s++ = ' ';
			t = fpp.address;
			while ( --num3 >= 0 ) *s++ = *t++;
			LineDealloc(Line);
			return(l);
		}
	}
	if ( ( num3 + LINEHEAD ) > Line->used ) {
/* 5-sep-1989 */
		t = Line->text + ( Line->used - LINEHEAD - 1 );
		s = Line->text + num3 - 1;
		i = num2;
		while ( --i >= 0 ) *s-- = *t--;
		i = virt1;
		while ( --i >= 0 ) *s-- = ' ';
		Line->used = num3 + LINEHEAD;
	}
	else {
		s = fp.address;
		i = virt1;
		while ( --i >= 0 ) *s++ = ' ';
		t = fpp.address;
		i = num2;
		while ( --i >= 0 ) *s++ = *t++;
		DropPart(Line,(WORD)(Line->used-num3-LINEHEAD));
	}
	return(Line);
}

/*
 		#] Dsline :
 		#[ InsText :

		Inserts Line2 inside Line1 at position 'to'.

*/

WORD mcharnum;

WORD
InsText(LINE Line1,WORD to,LINE Line2,LINE After)
{
	WORD virt, type;
	WORD i1, i2, n;
	UBYTE *s, *t;
	LINE l;
	FINDPOS fp;
	mcharnum = -1;
	if ( Line2->used < 0 ) {
		if ( ( to > 1 ) || ( Line1->used != LINEHEAD ) ) return((WORD)NoFold());
		return(CopyLine(Line2,After));
	}
	if ( Line2->used == LINEHEAD ) return( CopyLine(Line1,After) );
	if ( to == 1 ) {
		if ( Concat2(Line2,Line1,1,After) < 0 ) return(-1);
		mcharnum = Line2->used - LINEHEAD;
		return(0);
	}
	type = FindChar(Line1,to,&fp);
	virt = fp.virt;
	if ( ( type == INSIDETAB ) || ( type == ONTAB ) ) {
		if ( *(Line2->text) == '\t' ) virt = 0;
		i1 = fp.num-1;
	}
	else if ( type == NORMALCHAR ) i1 = fp.num-1;
	else i1 = fp.num;
	i2 = Line2->used - LINEHEAD;
	n = Line1->used - LINEHEAD + i2 + virt;
	if ( n > MAXINLINE ) {
		MesnOut(4);
		if ( MesAsk(98,INCOMMAND) <= 0 ) return(-1);
		/* "Truncate too long line? (Y/N) " */
		n = MAXINLINE;
	}
	if ( ( l = LineAlloc(n,After) ) == (LINE)0 ) return(NOMEMORY);
	s = l->text;
	t = Line1->text;
	while ( ( --i1 >= 0 ) && ( --n >= 0 ) ) *s++ = *t++;
	while ( ( --virt >= 0 ) && ( --n >= 0 ) ) *s++ = ' ';
	t = Line2->text;
	while ( ( --i2 >= 0 ) && ( --n >= 0 ) ) *s++ = *t++;
	mcharnum = (WORD)(s - l->text);
	t = fp.address;
	while ( --n >= 0 ) *s++ = *t++;
	return(0);
}

/*
 		#] InsText :
 		#[ ClearYank :

VOID
ClearYank(VOID)
{
	BufUnd(WhichYank);
}

 		#] ClearYank :
 		#[ Clear the paste buffer :

 		To be used when an action must be aborted.

*/

VOID
ClearPaste(VOID)
{
	BUFFER b = textbuffer[WhichYank];
	while ( b->first != (LINE)b ) Discard(b->first);
	inibuf(WhichYank);
}

/*
 		#] Clear the paste buffer :
  	#] Auxiliary paste routines :
  	#[ Kcopy :
*/

LONG
kcopy2(VOID)
{
	BUFFER b = cbuf;
	BUFFER y = textbuffer[WhichYank];
	LINE l;
	LINE first, last;
	WORD i, j = 0, indic = 0;
	WORD ii;
	WORD numstart, numlast, diffvirt, diffvcol, type;
	UBYTE *s;
	LONG RetVal = 0;
	if ( ( type = Range(&first,&last,&numstart,&numlast,
				&diffvirt,&diffvcol) ) < 0 ) return(-1L);
	BufUnd(WhichYank);
	CopSet(b->bufnum,WhichYank,NONAMES);	/* Put in the settings; */
	if ( type ) {
		if ( b->MarkVirt > 0 ) { indic = 1; diffvirt++; }
	}
	else {
		if ( b->newvirt > 0 ) { indic = 1; diffvirt++; }
	}
	if ( b->column > b->MarkColumn ) { i = b->MarkColumn; j = b->column; }
	else							 { j = b->MarkColumn; i = b->column; }

	while ( --diffvirt >= 0 ) {
		ii = j - i;
		if ( ( l = LineAlloc(ii,(LINE)y) ) == (LINE)0 ) goto NoSuccess;
		s = l->text;
		while ( --ii >= 0 ) *s++ = ' ';
	}

	if ( !indic ) {
		l = last;
		do {
			if ( Cpline(l,i,j,(LINE)y) ) {	/* used to not go to ClearUndo */
NoSuccess:
				ClearPaste();
				RetVal = NOMEMORYL;
				goto EndRet;
			}
		} while ( ( l = l->prev ) != first->prev );
	}
	y->current = y->first;
	y->totline = CountLines(y);
	y->newline = 1;
	y->newvirt = 0;
	y->column = 1;
EndRet:
	ClearUndo();
	return(RetVal);
}

LONG
kcopy(VOID)
{
	LONG ret;
	if ( commandmode ) return(0L);
	if ( ( ret = kcopy2() ) != 0 ) return(ret);
	if ( DispBlock <= 0 ) MesnOut(80); /* "Range copied." */
	return(0L);
}

/*
  	#] Kcopy :
  	#[ Copy :
  	*/

LONG
copy(VOID)
{
	BUFFER b = cbuf, y = textbuffer[WhichYank];
	LINE l, first, last;
	WORD i, j = 0, indic, Vindic = 0;
	WORD numstart, numlast, diffvirt, diffvcol, type;
	UBYTE *s;
	LONG RetVal = 0;
	if ( commandmode ) return(0L);
	if ( ( type = Range(&first,&last,&numstart,&numlast,
				&diffvirt,&diffvcol) ) < 0 ) return(-1L);
/*
 *	First copy the start of the first line, unless first and last are the same.
 */
	if ( ( first == last ) && ( diffvirt == 0 ) ) return(kcopy());
	BufUnd(WhichYank);
	CopSet(b->bufnum,WhichYank,NONAMES);	/* Put in the settings; */
	if ( type && ( b->newvirt > 0 ) ) { /* mark,cursor */
		i = b->column;
		j = diffvirt;
		Vindic = b->MarkVirt;
		indic = b->MarkColumn;
	}
	else if ( !type && ( b->MarkVirt > 0 ) ) { /* cursor,mark */
		i = b->MarkColumn;
		j = diffvirt;
		Vindic = b->newvirt;
		indic = b->column;
	}
	else {
		/* copy the start of the tail line */
		if ( type ) {
			i = b->column;
			indic = b->MarkColumn;
		}
		else {
			i = b->MarkColumn;
			indic = b->column;
		}
		if ( Cpline(last,1,i,(LINE)y) ) {
			RetVal = NOMEMORYL;
			goto EndRet;
		}
	}
	if ( j ) {	/* make j-1 empty lines and one with i-1 blanks. */
		j--;
		if ( i > 0 ) i--;
		if ( ( l = LineAlloc(i,(LINE)y) ) == (LINE)0 ) {
			RetVal = NOMEMORYL;
			goto EndRet;
		}
		s = l->text;
		while ( --i >= 0 ) *s++ = ' ';		/* There are the blanks */
		while ( --j >= 0 ) {
			if ( !LineAlloc(0,(LINE)y) ) goto NoSuccess;
		}
		if ( Vindic ) goto EndCopy;
		last = (LINE)b;
	}
	/* Copy the lines inbetween */
	l = last->prev;
	while ( l != first ) {
		if ( CopyLine(l,(LINE)y) ) goto NoSuccess;
		l = l->prev;
	}
	/* Copy the tail of the start line */
	if ( indic > 1 ) {
		if ( CopyToEnd(first,indic,(LINE)y) ) goto NoSuccess;
	}
	else {
		if ( CopyLine(first,(LINE)y) ) {
NoSuccess:
			ClearPaste();
			RetVal = NOMEMORYL;
			goto EndRet;
		}
	}

	/* Fix up the buffer */
EndCopy:
	y->current = y->first;
	y->totline = CountLines(y);
	y->newline = 1;
	y->newvirt = 0;
	y->column = 1;
	if ( DispBlock <= 0 ) MesnOut(80); /* "Range copied." */
EndRet:
	ClearUndo();
	return(RetVal);
}

/*
  	#] Copy :
  	#[ Kcut :
*/

LONG
kcut(VOID)
{
	LONG ret;
	BUFFER b = cbuf;
	LINE l;
	LINE first, last, before;
	WORD i, j = 0, indic = 0;
	WORD numstart, numlast, diffvirt, diffvcol, type;
	LONG virtlines;
	if ( commandmode ) return(0L);
	ClearUndo();

	if ( ( type = Range(&first,&last,&numstart,
			&numlast,&diffvirt,&diffvcol) ) < 0 ) return(-1L);
	l = first;
	before = l->prev;
	do { if ( l->used < 0 ) return(NoFold()); }
		while ( ( l = l->next ) != last->next );

	virtlines = b->newvirt;
	if ( ( ret = kcopy2() ) != 0 ) return(ret);

	if ( type ) { if ( b->MarkVirt > 0 ) { indic = 1; } }
	else		{ if ( b->newvirt  > 0 ) { indic = 1; } }

	if ( b->column > b->MarkColumn ) { i = b->MarkColumn; j = b->column; }
	else							 { j = b->MarkColumn; i = b->column; }

	if ( !indic ) {			/* now remove pieces. */
		l = last;
		do {
			l = Dsline(l,i,j);
		} while ( ( l = l->prev ) != before );
	}
	b->current = l->next;		/* Just incase Dsline changes a line */

	if ( type ) {
		b->column = b->MarkColumn;
		b->newvirt = b->MarkVirt;
		b->newline -= textbuffer[WhichYank]->totline - virtlines - 1;
	}
	b->mark = (LINE)b;			/* Delete the mark to prevent keyrepeat */
	b->MarkVirt = 0;
	b->dirtyflag = 1;
	ForceDisplay();
	if ( DispBlock <= 0 ) MesnOut(81); /* "Range cut." */
	return(0L);
}

/*
  	#] Kcut :
  	#[ Cut :
*/

LONG
cut(VOID)
{
	BUFFER b = cbuf;
	BUFFER y = textbuffer[WhichYank];
	LINE l;
	LINE first, last, before;
	WORD i, j = 0, indic, Vindic = 0;
	WORD numstart, numlast, diffvirt, diffvcol, type, screenlines;
	UBYTE *s;
	LONG RetVal = 0;
	LONG takenout = 0;
	if ( commandmode ) return(0L);
	ClearUndo();
	if ( ( type = Range(&first,&last,&numstart,&numlast,
				&diffvirt,&diffvcol) ) < 0 ) return(-1L);

		/* a single line can be dealt with by a different routine */

	if ( ( first == last ) && ( diffvirt == 0 ) ) return(kcut());

		/* Test for foldlines */

	if ( first->used < 0 ) {
		if ( ( type && ( b->MarkColumn > 1 ) ) ||
			 ( !type && ( b->column > 1 ) ) ) return(NoFold());
	}
	if ( last->used < 0 ) {
		if ( ( !type && ( b->MarkColumn > 1 ) ) ||
			 ( type && ( b->column > 1 ) ) ) return(NoFold());
	}

		/* Count screen lines. Help for updating screen. */

	l = first;	i = 0;
	while ( l != last ) {
		i++; l = l->next;
	}
	screenlines = i;

		/*	Final preparation: clear the yankbuffer, copy the settings and 
			keep the line in front of the entire range for when lines change */

	BufUnd(WhichYank);
	ClearUndo();
	CopSet(b->bufnum,WhichYank,NONAMES);	/* Put in the settings; */
	before = first->prev;

		/*	Get some parameters and test whether all is entirely virtual */

	if ( type && ( b->newvirt > 0 ) ) { /* mark,cursor */
		i = b->column;
		j = b->newvirt - b->MarkVirt;
		Vindic = b->MarkVirt;
		indic = b->MarkColumn;
	}
	else if ( !type && ( b->MarkVirt > 0 ) ) { /* cursor,mark */
		i = b->MarkColumn;
		j = b->MarkVirt - b->newvirt;
		Vindic = b->newvirt;
		indic = b->column;
	}
	else {			/* copy the start of the tail line */
		if ( type ) {
			i = b->column;
			indic = b->MarkColumn;
		}
		else {
			i = b->MarkColumn;
			indic = b->column;
		}
		if ( Cpline(last,1,i,(LINE)y) ) {
			RetVal = NOMEMORYL;
			goto EndRet;
		}
	}
	if ( j ) {		/* make j-1 empty lines and one with i blanks. */
		j--;
		if ( i > 0 ) i--;
		if ( ( l = LineAlloc(i,(LINE)y) ) == (LINE)0 ) {
			RetVal = NOMEMORYL;
			goto EndRet;
		}
		s = l->text;
		while ( --i >= 0 ) *s++ = ' ';
		while ( --j >= 0 ) {
			if ( !LineAlloc(0,(LINE)y) ) { ClearPaste();
			RetVal = NOMEMORYL;
			goto EndRet;
			}
		}
		if ( Vindic ) goto EndCut;		/* only virtual territory */
		last = (LINE)b;
	}

		/* Copy the tail of the start line */

	if ( type ) { i = b->MarkColumn; j = b->column; }
	else		{ j = b->MarkColumn; i = b->column; }
	if ( first->used < 0 ) {		/* start line is foldline -> move */
		takenout += LINESINFOLD(first);
		l = first->prev;
		first->prev->next = first->next;
		first->next->prev = first->prev;
		first->next = y->first;
		first->prev = (LINE)y;
		y->first->prev = first;
		y->first = first;
		first = l;
		if ( last != (LINE)b && last->used > 0 ) last = Dsline(last,1,j);
	}
	else {
		if ( CopyToEnd(first,indic,(LINE)y) ) { ClearPaste();
			RetVal = NOMEMORYL;
			goto EndRet;
		}
		if ( last != (LINE)b ) {
			if ( last->used > 0 ) {
				if ( Concat3(first,i,last,j,last) ) {
					ClearPaste();
					RetVal = NOMEMORYL;
					goto EndRet;
				}
				first = first->prev;
				last = last->next;
				LineDealloc(first->next);
				LineDealloc(last->prev);
				takenout++;
			}
			else {
				first = first->prev;
				LineDealloc(first->next);
				takenout++;
			}
		}
		else {
			if ( first->used > LINEHEAD ) {
				j = FindColumn(first,(first->used-LINEHEAD)) + 1;
				if ( j > i ) first = Dsline(first,i,j);
			}
		}
	}

		/* Move now the lines inbetween */

	if ( first->next != last ) {
		takenout += NumLines(first->next,last);
		l = y->first->next;
		first->next->prev = y->first;
		y->first->next = first->next;
		last->prev->next = l;
		l->prev = last->prev;
		first->next = last;
		last->prev = first;
	}

		/* Fix up the buffers */

EndCut:
	y->current = y->first;
	y->totline = CountLines(y);
	y->newline = 1;
	y->newvirt = 0;
	y->column = 1;
	b->totline -= takenout;
	if ( type ) {
		b->column = b->MarkColumn;
		if ( ( b->newvirt = b->MarkVirt ) == 0 ) { b->newline -= takenout; }
	}
	/*	The ifs are for the difficult exceptional cases. This avoids
		the need to try to intercept them directly. */
	b->current = before->next;
	if ( b->current == (LINE)b ) { b->current = b->last; b->newvirt = 1; }
	if ( b->newline > b->totline ) { b->newline = b->totline; b->newvirt = 1; }
	if ( type ) {
		b->oldline = b->newline;
		b->screenline -= screenlines;
	}
	b->dirtyflag = 1;
	b->mark = (LINE)b;		/* Erase the mark. This was deemed necessary */
	b->MarkVirt = 0;		/* to avoid the loss of data when the user   */
	ForceDisplay();				/* presses F4 after the cut instead of F5    */
	if ( DispBlock <= 0 ) MesnOut(81); /* "Range cut." */
EndRet:
	ClearUndo();
	return(RetVal);
}

/*
  	#] Cut :
  	#[ kpaste :
*/

LONG
kpaste(VOID)
{
	BUFFER b = cbuf, oldbuf;
	BUFFER y = textbuffer[WhichYank];
	LINE l;
	LINE lb, la, ly;
	WORD i, j, ii, jj;
	LONG NumInY;
	struct tline EmpLine;
	LINE EmptyLine;
	WORD Oldmode, markset = 0;
	if ( commandmode == 2 ) commandmode = 1;
	Oldmode = commandmode;
	NumInY = y->totline;
	if ( NumInY <= 0 ) return(0L);
	if ( commandmode ) { NumInY = 1; b = commandbuffer; }
	l = y->first;
	do { if ( l->used < 0 ) return(NoFold()); }		/* Fold in yankbuffer */
				while ( ( l = l->next ) != (LINE)y );

	ClearUndo();
	l = b->current;
	lb = l->prev;
	j = (WORD)NumInY;
	if ( b->newvirt <= 0 ) {
		do {
			if ( l->used < 0 ) return(NoFold());	/* Fold in pasterange */
			j--;
		} while ( ( ( l = l->next ) != (LINE)b ) && ( j > 0 ) );
		l = l->prev;
		i = (WORD)NumInY - j;		/* i = number of insertions */
		la = b->current;	/* j = virtual lines */
		lb = b->current->prev;
	}
	else {
		while ( b->newvirt > 1 ) {
			if ( ( l = LineAlloc(0,b->last) ) == (LINE)0 ) return(NOMEMORYL);
			b->current = l;
			b->newline++;
			b->totline++;
			b->newvirt--;
		}
		i = 0; j = (WORD)NumInY; la = (LINE)b;
		lb = b->current;
	}

/*\
 *	Procedure:	Make j new lines with b->column leading blanks.
 *				Then add b->newvirt-1 empty lines or make i insertions.
 *
 *				Note that we run backwards. This makes insertions easier.
 *
 *			In case of trouble we delete the lines lb->next to la->prev incl.
 *			If successful we delete ii lines, starting at b->current;
\*/

	ly = y->last;
	if ( commandmode ) ly = y->first;
	jj = j;
	commandmode = 0;
	if ( j > 0 ) {
		EmptyLine = &EmpLine;
		EmptyLine->used = LINEHEAD;
		do {
			if ( Concat1(EmptyLine,b->column,ly,lb) ) goto AbortKpaste;
			if ( !markset ) {
				b->mark = lb->next;
				markset = 1;
				b->MarkChar = b->mark->used - LINEHEAD;
				b->MarkVoff = ( b->MarkChar > 0 ) ? 1: 0;
				b->MarkVirt = 0;
				DetMarkColumn();
			}
			ly = ly->prev;
		} while ( --j > 0 );
	}

	ii = i;
	while ( --i >= 0 ) {
		oldbuf = cbuf;
		cbuf = b;
		if ( InsText(l,b->column,ly,lb) ) {
			cbuf = oldbuf;
			goto AbortKpaste;
		}
		cbuf = oldbuf;
		if ( !markset ) {
			b->mark = lb->next;
			markset = 1;
			if ( mcharnum < 0 ) {
				b->MarkColumn = b->column;
			}
			else {
				b->MarkChar = mcharnum;
				b->MarkVoff = ( b->MarkChar > 0 ) ? 1: 0;
				b->MarkVirt = 0;
				DetMarkColumn();
			}
		}
		ly = ly->prev;
		l = l->prev;
	}

	i = ii;
	l = b->current->prev;
	while ( --i >= 0 ) LineDealloc(l->next);
	commandmode = Oldmode;
/*
 *	Finally recompute the screen and show it.
 */
	if ( ii > 0 ) b->current = lb->next;
	else {
		while ( b->newvirt > 0 ) {
			b->current = b->current->next;
			b->newline++;
			b->newvirt--;
		}
	}
	b->totline += jj;
	b->dirtyflag = 1;
	if ( commandmode ) MessageCommand();
	else               ForceDisplay();
	return(0L);

AbortKpaste:
	commandmode = Oldmode;
	la = la->prev;
	l = lb;
	while ( l != la ) LineDealloc(l->next);
	return(NOMEMORYL);
}

/*
  	#] kpaste :
  	#[ Paste :

	No code for the mark yet

*/

LONG
paste(VOID)
{
	BUFFER b = cbuf;
	BUFFER y = textbuffer[WhichYank];
	struct tline EmpLine;
	LINE EmptyLine, la, ly;
	LINE lb;
	WORD j, jj;
	WORD oldcol;
	WORD WasFold = 0;
	UBYTE blank = ' ';
	if ( y->totline == 0 ) return(0L);
	if ( commandmode || ( y->totline == 1 ) ) return(kpaste());
	if ( ( y->first->used < 0 ) && ( b->column != 1 ) ) return(NoFold());
	ClearUndo();
	if ( b->totline == 0 ) {	/* empty buffer. Force to become real */
		oldcol = b->column;
		if ( Cinsert(&blank,1) ) return(NOMEMORYL);
		b->column = oldcol;
		Tdelete();
	}
	lb = b->current;
	la = lb->next;
	ly = y->last;
	if ( b->newvirt > 0 ) {
		b->mark = lb->next;
		while ( ly->prev != (LINE)y ) {
			if ( CopyLine(ly,lb) ) goto AbortPaste;
			ly = ly->prev;
		}
		if ( ly->used < 0 ) {
			if ( CopyLine(ly,lb) ) goto AbortPaste;
		}
		else {
			EmptyLine = &EmpLine;
			EmptyLine->used = LINEHEAD;
			if ( Concat1(EmptyLine,b->column,ly,lb) ) goto AbortPaste;
		}
		ly = lb->next;
		jj = j = b->newvirt - 1;
		while ( --j >= 0 ) if ( LineAlloc(0,lb) == (LINE)0 ) goto AbortPaste;
		b->totline += jj + y->totline;
		b->newline += jj + 1;
		b->current = ly;
		b->newvirt = 0;
		b->mark = b->last;
		b->MarkChar = b->mark->used - LINEHEAD;
		b->MarkVoff = ( b->MarkChar > 0 ) ? 1: 0;
		b->MarkVirt = 0;
		DetMarkColumn();
	}
	else {
		if ( lb->used < 0 ) {
			if ( ( ly->used != LINEHEAD ) || ( b->column != 1 ) )
												return((WORD)(NoFold()));
			WasFold = 1;
			lb = lb->prev;
		}
		else if ( ( j = Concat2(ly,lb,b->column,lb) ) < 0 ) return(NOMEMORYL);
		b->mark = lb->next;
		b->MarkVirt = 0;
		b->MarkColumn = 1;
		ly = ly->prev;
		while ( ly->prev != (LINE)y ) {
			if ( CopyLine(ly,lb) ) goto AbortPaste;
			ly = ly->prev;
		}
		if ( WasFold || ly->used < 0 ) {
			if ( CopyLine(ly,lb) ) goto AbortPaste;
		}
		else if ( Concat1(lb,b->column,ly,lb) ) goto AbortPaste;
		b->current = lb->next;
		if ( !WasFold ) LineDealloc(lb);
		b->totline += y->totline - 1;
		DetMark();
	}
	b->dirtyflag = 1;
	ForceDisplay();
	return(0L);

AbortPaste:
	la = la->prev;
	if ( WasFold ) lb = lb->next;
	while ( lb != la ) {
		ly = la;
		la = la->prev;
		if ( ly->used >= 0 ) LineDealloc(ly);
		else                 RemoveFold(ly);
	}
	return(NOMEMORYL);
}

/*
  	#] Paste :
*/

